import logging
import socket
import sqlite3
from abc import ABC, abstractmethod
from datetime import datetime
from os import path
from pathlib import Path
from typing import List, Dict, Union
from uuid import uuid4

import numpy as np
import pandas as pd
from numpy import nan_to_num

from aequilibrae import Parameters
from aequilibrae.context import get_active_project
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.paths.linear_approximation import LinearApproximation
from aequilibrae.paths.optimal_strategies import OptimalStrategies
from aequilibrae.paths.traffic_class import TrafficClass, TransportClassBase
from aequilibrae.paths.vdf import VDF, all_vdf_functions
from aequilibrae.project.database_connection import database_connection
from aequilibrae.utils.core_setter import set_cores


class AssignmentBase(ABC):
    def __init__(self, project=None):
        self.procedure_id = uuid4().hex
        self.procedure_date = str(datetime.today())

        proj = project or get_active_project(must_exist=False)
        self.project = proj

        self.parameters = proj.parameters if proj else Parameters().parameters
        self.logger = proj.logger if proj else logging.getLogger("aequilibrae")

        self.classes: List[TrafficClass] = []
        self.algorithm: str = None
        self.time_field: str = None
        self.assignment: Union[LinearApproximation, OptimalStrategies] = None
        self.free_flow_tt: np.ndarray = None
        self.total_flow: np.ndarray = None
        self.cores: int = None
        self._config = {}

        self.description: str = ""

    def algorithms_available(self) -> list:
        """
        Returns all algorithms available for use

        :Returns:
            :obj:`list`: List of string values to be used with **set_algorithm**
        """
        return self.all_algorithms

    @abstractmethod
    def set_algorithm(self, algorithm: str):
        pass

    @abstractmethod
    def set_cores(self, cores: int) -> None:
        pass

    def execute(self, log_specification=True) -> None:
        """Processes assignment"""
        if log_specification:
            self.log_specification()
        self.assignment.execute()

    @abstractmethod
    def log_specification(self):
        pass

    @abstractmethod
    def save_results(self, table_name: str, keep_zero_flows=True, project=None) -> None:
        pass

    @abstractmethod
    def results(self) -> pd.DataFrame:
        pass

    def report(self) -> pd.DataFrame:
        """Returns the assignment convergence report

        :Returns:
           **DataFrame** (:obj:`pd.DataFrame`): Convergence report
        """
        return pd.DataFrame(self.assignment.convergence_report)

    @abstractmethod
    def info(self) -> dict:
        pass

    def set_classes(self, classes: List[TransportClassBase]) -> None:
        """
        Sets Transport classes to be assigned

        :Arguments:
            **classes** (:obj:`List[TransportClassBase]`): List of TransportClass's for assignment
        """

        ids = {x._id for x in classes}
        if len(ids) < len(classes):
            raise ValueError("Classes need to be unique. Your list of classes has repeated items/IDs")
        self.classes = classes  # type: List[TransportClassBase]

    def add_class(self, transport_class: TransportClassBase) -> None:
        """
        Adds a Transport class to the assignment

        :Arguments:
            **transport_class** (:obj:`TransportClassBase`): Transport class
        """

        ids = [x._id for x in self.classes if x._id == transport_class._id]
        if len(ids) > 0:
            raise ValueError("Transport class already in the assignment")

        self.classes.append(transport_class)

    def _check_field(self, field: str) -> None:
        """Throws expection if field is invalid."""
        if not self.classes:
            raise ValueError("You need add at least one transport class first")

        for c in self.classes:
            if field not in c.graph.graph.columns:
                raise ValueError(f"'{field}' not in graph for '{c._id}'")

            if np.any(np.isnan(c.graph.graph[field].values)):
                raise ValueError(f"At least one link for {field} is NaN for '{c._id}'")

            if c.graph.graph[field].values.min() <= 0:
                raise ValueError(f"There is at least one link with zero or negative {field} for '{c._id}'")

    def set_time_field(self, time_field: str) -> None:
        self._check_field(time_field)
        c = self.classes[0]
        self.free_flow_tt = np.zeros(c.graph.graph.shape[0], c.graph.default_types("float"))
        self.free_flow_tt[c.graph.graph.__supernet_id__] = c.graph.graph[time_field]
        self.total_flow = np.zeros(self.free_flow_tt.shape[0], np.float64)
        self.time_field = time_field


class TrafficAssignment(AssignmentBase):
    """Traffic assignment class.

    For a comprehensive example on use, see the Use examples page.

    .. code-block:: python

        >>> from aequilibrae.paths import TrafficAssignment, TrafficClass

        >>> project = create_example(project_path)
        >>> project.network.build_graphs()

        >>> graph = project.network.graphs['c'] # we grab the graph for cars
        >>> graph.set_graph('free_flow_time') # let's say we want to minimize time
        >>> graph.set_skimming(['free_flow_time', 'distance']) # And will skim time and distance
        >>> graph.set_blocked_centroid_flows(True)

        >>> proj_matrices = project.matrices

        >>> demand = proj_matrices.get_matrix("demand_omx")

        # We will only assign one user class stored as 'matrix' inside the OMX file
        >>> demand.computational_view(['matrix'])

        # Creates the assignment class
        >>> assigclass = TrafficClass("car", graph, demand)

        >>> assig = TrafficAssignment()

        # The first thing to do is to add at list of traffic classes to be assigned
        >>> assig.set_classes([assigclass])

        # Then we set the volume delay function
        >>> assig.set_vdf("BPR")  # This is not case-sensitive

        # And its parameters
        >>> assig.set_vdf_parameters({"alpha": "b", "beta": "power"})

        # The capacity and free flow travel times as they exist in the graph
        >>> assig.set_capacity_field("capacity")
        >>> assig.set_time_field("free_flow_time")

        # And the algorithm we want to use to assign
        >>> assig.set_algorithm('bfw')

        >>> assig.max_iter = 10
        >>> assig.rgap_target = 0.00001

        >>> assig.execute() # we then execute the assignment

        # If you want, it is possible to access the convergence report
        >>> convergence_report = pd.DataFrame(assig.assignment.convergence_report)

        # Assignment results can be viewed as a Pandas DataFrame
        >>> results_df = assig.results()

        # Information on the assignment setup can be recovered with
        >>> info = assig.info()

        # Or save it directly to the results database
        >>> results = assig.save_results(table_name='base_year_assignment')

        # skims are here
        >>> avg_skims = assigclass.results.skims # blended ones
        >>> last_skims = assigclass._aon_results.skims # those for the last iteration
    """

    bpr_parameters = ["alpha", "beta"]
    all_algorithms = ["all-or-nothing", "msa", "frank-wolfe", "fw", "cfw", "bfw"]

    def __init__(self, project=None) -> None:
        """"""
        self.__dict__["_TrafficAssignment__initalised"] = False
        super().__init__(project=project)

        proj = project or get_active_project(must_exist=False)

        par = proj.parameters if proj else Parameters().parameters
        parameters = par["assignment"]["equilibrium"]

        self.rgap_target = parameters["rgap"]
        self.max_iter = parameters["maximum_iterations"]
        self.vdf = VDF()
        self.vdf_parameters = None  # type: list
        self.capacity_field = None  # type: str
        self.capacity = None  # type: np.ndarray
        self.congested_time = None  # type: np.ndarray
        self.save_path_files = False  # type: bool
        self.preloads = None  # type: pd.DataFrame

        self.steps_below_needed_to_terminate = 1

        self._config = {}

        self.__initalised = True

    def __setattr__(self, name, value) -> None:
        # Special methods (__method__) cannot be override at runtime, instead we'll
        # just set a variable to indicate if the checking of attributes should be performed
        if self.__initalised:
            check, value, message = self.__check_attributes(name, value)
            if not check:
                raise ValueError(message)
        super().__setattr__(name, value)

    def __check_attributes(self, instance, value):
        if instance == "rgap_target":
            if not isinstance(value, float):
                return False, value, "Relative gap needs to be a float"
            if isinstance(self.assignment, LinearApproximation):
                self.assignment.rgap_target = value
        elif instance == "max_iter":
            if not isinstance(value, int):
                return False, value, "Number of iterations needs to be an integer"
            if isinstance(self.assignment, LinearApproximation):
                self.assignment.max_iter = value
        elif instance == "vdf":
            v = value.lower()
            if v not in all_vdf_functions:
                return False, value, f"Volume-delay function {value} is not available"
            value = VDF()
            value.function = v
        elif instance == "classes":
            if isinstance(value, TrafficClass):
                value = [value]
            elif isinstance(value, list):
                for v in value:
                    if not isinstance(v, TrafficClass):
                        return False, value, "Traffic classes need to be proper AssignmentClass objects"
            else:
                raise ValueError("Traffic classes need to be proper AssignmentClass objects")
        elif instance == "vdf_parameters":
            if not self.__validate_parameters(value):
                return False, value, f"Parameter set is not valid: {value} "
        elif instance in ["time_field", "capacity_field"] and not isinstance(value, str):
            return False, value, f"Value for {instance} is not string"
        elif instance == "cores" and not isinstance(value, int):
            return False, value, f"Value for {instance} is not integer"
        elif instance == "save_path_files" and not isinstance(value, bool):
            return False, value, f"Value for {instance} is not boolean"
        if instance not in self.__dict__:
            return False, value, f"TrafficAssignment class does not have property {instance}"
        return True, value, ""

    def set_vdf(self, vdf_function: str) -> None:
        """
        Sets the Volume-delay function to be used

        :Arguments:
            **vdf_function** (:obj:`str`): Name of the VDF to be used
        """
        self.vdf = vdf_function

    def set_classes(self, classes: List[TrafficClass]) -> None:
        """
        Sets Traffic classes to be assigned

        :Arguments:
            **classes** (:obj:`List[TrafficClass]`): List of Traffic classes for assignment
        """

        ids = {x._id for x in classes}
        if len(ids) < len(classes):
            raise ValueError("Classes need to be unique. Your list of classes has repeated items/IDs")
        self.classes = classes  # type: List[TrafficClass]

    def add_class(self, traffic_class: TrafficClass) -> None:
        """
        Adds a traffic class to the assignment

        :Arguments:
            **traffic_class** (:obj:`TrafficClass`): Traffic class
        """

        ids = [x._id for x in self.classes if x._id == traffic_class._id]
        if len(ids) > 0:
            raise ValueError("Traffic class already in the assignment")

        self.classes.append(traffic_class)

    # TODO: Create procedure to check that travel times, capacities and vdf parameters are equal across all graphs
    # TODO: We also need procedures to check that all graphs are compatible (i.e. originated from the same network)
    def set_algorithm(self, algorithm: str):
        """
        Chooses the assignment algorithm. e.g. 'frank-wolfe', 'bfw', 'msa'

        'fw' is also accepted as an alternative to 'frank-wolfe'

        :Arguments:
            **algorithm** (:obj:`str`): Algorithm to be used
        """

        # First we instantiate the arrays we will be using over and over

        algo_dict = {i: i for i in self.all_algorithms}
        algo_dict["fw"] = "frank-wolfe"
        algo = algo_dict.get(algorithm.lower())

        if algo is None:
            raise AttributeError(f"Assignment algorithm not available. Choose from: {','.join(self.all_algorithms)}")

        if algo in ["all-or-nothing", "msa", "frank-wolfe", "cfw", "bfw"]:
            self.assignment = LinearApproximation(self, algo, project=self.project)
        else:
            raise ValueError("Algorithm not listed in the case selection")

        self.__dict__["algorithm"] = algo
        self._config["Algorithm"] = algo
        self._config["Maximum iterations"] = self.assignment.max_iter
        self._config["Target RGAP"] = self.assignment.rgap_target

    def set_vdf_parameters(self, par: dict) -> None:
        """
        Sets the parameters for the Volume-delay function.

        Parameter values can be scalars (same values for the entire network) or network field names
        (link-specific values) - Examples: {'alpha': 0.15, 'beta': 4.0} or  {'alpha': 'alpha', 'beta': 'beta'}

        :Arguments:
            **par** (:obj:`dict`): Dictionary with all parameters for the chosen VDF

        """
        if self.classes is None or self.vdf.function.lower() not in all_vdf_functions:
            raise RuntimeError(
                "Before setting vdf parameters, you need to set traffic classes and choose a VDF function"
            )
        self.__dict__["vdf_parameters"] = par
        self._config["VDF parameters"] = par
        pars = []
        if self.vdf.function in ["BPR", "BPR2", "CONICAL", "INRETS"]:
            for p1 in ["alpha", "beta"]:
                if p1 not in par:
                    raise ValueError(f"{p1} should exist in the set of parameters provided")
                p = par[p1]
                if isinstance(self.vdf_parameters[p1], str):
                    c = self.classes[0]
                    array = np.zeros(c.graph.graph.shape[0], c.graph.default_types("float"))
                    array[c.graph.graph.__supernet_id__] = c.graph.graph[p]
                else:
                    array = np.zeros(self.classes[0].graph.graph.shape[0], np.float64)
                    array.fill(self.vdf_parameters[p1])
                pars.append(array)

                if np.any(np.isnan(array)):
                    raise ValueError(f"At least one {p1} is NaN")

                if p1 == "alpha":
                    if array.min() < 0:
                        raise ValueError(f"At least one {p1} is smaller than zero")
                else:
                    if array.min() < 1:
                        raise ValueError(f"At least one {p1} is smaller than one. Results will make no sense")

        self.__dict__["vdf_parameters"] = pars
        self._config["VDF function"] = self.vdf.function.lower()

    def set_cores(self, cores: int) -> None:
        """Allows one to set the number of cores to be used AFTER traffic classes have been added

        Inherited from :obj:`AssignmentResultsBase`

        :Arguments:
            **cores** (:obj:`int`): Number of CPU cores to use
        """
        if not self.classes:
            raise RuntimeError("You need load traffic classes before overwriting the number of cores")

        self.cores = set_cores(cores)
        for c in self.classes:
            c.results.set_cores(self.cores)
            c._aon_results.set_cores(self.cores)

    def set_save_path_files(self, save_it: bool) -> None:
        """Turn path saving on or off.

        :Arguments:
            **save_it** (:obj:`bool`): Boolean to indicate whether paths should be saved
        """
        if self.classes is None:
            raise RuntimeError("You need to set traffic classes before turning path saving on or off")

        # self.save_path_files = save_it
        for c in self.classes:
            c._aon_results.save_path_file = save_it

    def set_path_file_format(self, file_format: str) -> None:
        """Specify path saving format. Either parquet or feather.

        :Arguments:
            **file_format** (:obj:`str`): Name of file format to use for path files
        """
        if self.classes is None:
            raise RuntimeError("You need to set traffic classes before specifying path saving options")

        if file_format == "feather":
            for c in self.classes:
                c._aon_results.write_feather = True
        elif file_format == "parquet":
            for c in self.classes:
                c._aon_results.write_feather = False
        else:
            raise TypeError(f"Unsupported path file format {file_format} - only feather or parquet available.")

    def set_time_field(self, time_field: str) -> None:
        """
        Sets the graph field that contains free flow travel time -> e.g. 'fftime'

        :Arguments:
            **time_field** (:obj:`str`): Field name
        """

        super().set_time_field(time_field)

        self.__dict__["congested_time"] = np.array(self.free_flow_tt, copy=True)
        self._config["Time field"] = time_field

    def set_capacity_field(self, capacity_field: str) -> None:
        """
        Sets the graph field that contains link capacity for the assignment period -> e.g. 'capacity1h'

        :Arguments:
            **capacity_field** (:obj:`str`): Field name
        """
        super()._check_field(capacity_field)
        c = self.classes[0]

        self.cores = c.results.cores
        self.capacity = np.zeros(c.graph.graph.shape[0], c.graph.default_types("float"))
        self.capacity[c.graph.graph.__supernet_id__] = c.graph.graph[capacity_field]
        self.capacity_field = capacity_field
        self._config["Number of cores"] = c.results.cores
        self._config["Capacity field"] = capacity_field

    def add_preload(self, preload: pd.DataFrame, name: str = None) -> None:
        """
        Given a dataframe of 'link_id', 'direction' and 'preload', merge into current preloads dataframe.

        :Arguments:
            **preload** (:obj:`pd.DataFrame`): dataframe mapping 'link_id' & 'direction' to 'preload'
            **name** (:obj:`str`): Name for particular preload (optional - default name will be chosen if not specified)
        """
        # Create preloads dataframe in correct order if not already initialised
        if self.preloads is None:
            g = self.classes[0].graph.graph
            self.preloads = g.sort_values(by="__supernet_id__")[["link_id", "direction"]].copy()

        # Check that columns of preload are link_id, direction, preload:
        expected = {"link_id", "direction", "preload"}
        missing = expected - set(preload.columns)
        additional = set(preload.columns) - expected
        if missing:
            raise ValueError(
                f"Input preload dataframe is missing columns: {missing}\n" f"expected columns are {expected}"
            )
        elif additional:
            raise ValueError(
                f"Input preload dataframe has additional columns: {additional}\n" f"expected columns are {expected}"
            )

        # Reject empty preloads
        if len(preload) == 0:
            raise ValueError("Cannot set empty preload!")

        # Check name is not already used (generate new name if needed):
        name = (
            name if name else f"preload_{len(self.preloads.columns) - 1}"
        )  # -1 -> remove keys to get 1 indexed preload columns
        if name in self.preloads.columns:
            raise ValueError(f"New preload has duplicate name - already used names are: {self.preload.columns}")
        preload.rename(columns={"preload": name}, inplace=True)

        # Merge onto current preload dataframe
        self.preloads = pd.merge(self.preloads, preload, on=["link_id", "direction"], how="left")
        self.preloads[name] = self.preloads[name].fillna(0)

        # Enable preload to be added before or after specifyig the algorithm
        if self.assignment is not None:
            if self.assignment.preload is None:
                self.assignment.preload = self.preloads[name].to_numpy()
            else:
                self.assignment.preload += self.preloads[name]

    # TODO: This function actually needs to return a human-readable dictionary, and not one with
    #       tons of classes. Feeds into the class above
    # def load_assignment_spec(self, specs: dict) -> None:
    #     pass
    # def get_spec(self) -> dict:
    #     """Gets the entire specification of the assignment"""
    #     return deepcopy(self.__dict__)

    def __validate_parameters(self, kwargs) -> bool:
        if self.vdf == "":
            raise ValueError("First you need to set the Volume-Delay Function to use")

        par = list(kwargs.keys())
        q = [x for x in par if x not in self.bpr_parameters] + [x for x in self.bpr_parameters if x not in par]
        if len(q) > 0:
            raise ValueError("List of functions {} for vdf {} has an inadequate set of parameters".format(q, self.vdf))
        return True

    def log_specification(self):
        self.logger.info("Traffic Class specification")
        for cls in self.classes:
            self.logger.info(str(cls.info))

        self.logger.info("Traffic Assignment specification")
        self.logger.info(self._config)

    def save_results(self, table_name: str, keep_zero_flows=True, project=None) -> None:
        """Saves the assignment results to results_database.sqlite

        Method fails if table exists

        :Arguments:
            **table_name** (:obj:`str`): Name of the table to hold this assignment result

            **keep_zero_flows** (:obj:`bool`): Whether we should keep records for zero flows. Defaults to ``True``

            **project** (:obj:`Project`, *Optional*): Project we want to save the results to.
            Defaults to the active project
        """

        df = self.results()
        if not keep_zero_flows:
            df = df[df.PCE_tot > 0]

        if not project:
            project = self.project or get_active_project()
        conn = sqlite3.connect(path.join(project.project_base_path, "results_database.sqlite"))
        df.to_sql(table_name, conn)
        conn.close()

        conn = project.connect()
        report = {"convergence": str(self.assignment.convergence_report), "setup": str(self.info())}
        data = [table_name, "traffic assignment", self.procedure_id, str(report), self.procedure_date, self.description]
        conn.execute(
            """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                            description) Values(?,?,?,?,?,?)""",
            data,
        )
        conn.commit()
        conn.close()

    def results(self) -> pd.DataFrame:
        """Prepares the assignment results as a Pandas DataFrame

        :Returns:
            **DataFrame** (:obj:`pd.DataFrame`): Pandas DataFrame with all the assignment results indexed on `link_id`
        """

        idx = self.classes[0].graph.graph.__supernet_id__
        assig_results = [cls.results.get_load_results() for cls in self.classes]

        class1 = self.classes[0]
        res1 = assig_results[0]

        tot_flow = self.assignment.fw_total_flow[idx]
        voc = tot_flow / self.capacity[idx]
        congested_time = self.congested_time[idx]
        free_flow_tt = self.free_flow_tt[idx]
        preload = np.full(len(tot_flow), np.nan) if self.assignment.preload is None else self.assignment.preload

        fields = [
            "Preload_AB",
            "Preload_BA",
            "Preload_tot",
            "Congested_Time_AB",
            "Congested_Time_BA",
            "Congested_Time_Max",
            "Delay_factor_AB",
            "Delay_factor_BA",
            "Delay_factor_Max",
            "VOC_AB",
            "VOC_BA",
            "VOC_max",
            "PCE_AB",
            "PCE_BA",
            "PCE_tot",
        ]

        agg = pd.DataFrame([], columns=fields, index=res1.index[:]).astype(float)
        agg.fillna(0.0, inplace=True)

        # Use the first class to get a graph -> network link ID mapping
        m = class1.results.get_graph_to_network_mapping()
        graph_ab, graph_ba = m.graph_ab_idx, m.graph_ba_idx
        agg["Preload_AB"].values[m.network_ab_idx] = nan_to_num(preload[m.graph_ab_idx])
        agg["Preload_BA"].values[m.network_ba_idx] = nan_to_num(preload[m.graph_ba_idx])
        agg.loc[:, "Preload_tot"] = np.nansum([agg.Preload_AB, agg.Preload_BA], axis=0)

        agg["Congested_Time_AB"].values[m.network_ab_idx] = nan_to_num(congested_time[m.graph_ab_idx])
        agg["Congested_Time_BA"].values[m.network_ba_idx] = nan_to_num(congested_time[m.graph_ba_idx])
        agg.loc[:, "Congested_Time_Max"] = np.nanmax([agg.Congested_Time_AB, agg.Congested_Time_BA], axis=0)

        agg["Delay_factor_AB"].values[m.network_ab_idx] = nan_to_num(congested_time[graph_ab] / free_flow_tt[graph_ab])
        agg["Delay_factor_BA"].values[m.network_ba_idx] = nan_to_num(congested_time[graph_ba] / free_flow_tt[graph_ba])
        agg.loc[:, "Delay_factor_Max"] = np.nanmax([agg.Delay_factor_AB, agg.Delay_factor_BA], axis=0)

        agg["VOC_AB"].values[m.network_ab_idx] = nan_to_num(voc[m.graph_ab_idx])
        agg["VOC_BA"].values[m.network_ba_idx] = nan_to_num(voc[m.graph_ba_idx])
        agg.loc[:, "VOC_max"] = np.nanmax([agg.VOC_AB, agg.VOC_BA], axis=0)

        agg["PCE_AB"].values[m.network_ab_idx] = nan_to_num(tot_flow[m.graph_ab_idx])
        agg["PCE_BA"].values[m.network_ba_idx] = nan_to_num(tot_flow[m.graph_ba_idx])
        agg.loc[:, "PCE_tot"] = np.nansum([agg.PCE_AB, agg.PCE_BA], axis=0)

        assig_results.append(agg)
        return pd.concat(assig_results, axis=1).rename_axis("link_id")

    def info(self) -> dict:
        """Returns information for the traffic assignment procedure

        Dictionary contains keys  'Algorithm', 'Classes', 'Computer name', 'Procedure ID',
        'Maximum iterations' and 'Target RGap'.

        The classes key is also a dictionary with all the user classes per traffic class and their respective
        matrix totals

        :Returns:
            **info** (:obj:`dict`): Dictionary with summary information
        """

        classes = {}

        for cls in self.classes:
            uclass = {}

            if len(cls.matrix.view_names) == 1:
                uclass["matrix_totals"] = {nm: np.sum(cls.matrix.matrix_view[:, :]) for nm in cls.matrix.view_names}
            else:
                uclass["matrix_totals"] = {
                    nm: np.sum(cls.matrix.matrix_view[:, :, i]) for i, nm in enumerate(cls.matrix.view_names)
                }
            uclass["network mode"] = cls.graph.mode
            uclass["Value-of-time"] = cls.vot
            uclass["PCE"] = cls.pce
            if cls.fixed_cost_field:
                uclass["Fixed cost field"] = cls.fixed_cost_field
                uclass["Fixed cost multiplier"] = cls.fc_multiplier
            uclass["save_path_files"] = cls._aon_results.save_path_file
            uclass["path_file_feather_format"] = cls._aon_results.write_feather

            classes[cls._id] = uclass

        info = {
            "Algorithm": self.algorithm,
            "Classes": classes,
            "Computer name": socket.gethostname(),
            "Maximum iterations": self.assignment.max_iter,
            "Procedure ID": self.procedure_id,
            "Target RGap": self.assignment.rgap_target,
        }
        return info

    def save_skims(self, matrix_name: str, which_ones="final", format="omx", project=None) -> None:
        """Saves the skims (if any) to the skim folder and registers in the matrix list

        :Arguments:
            **name** (:obj:`str`): Name of the matrix record to hold this matrix (same name used for file name)

            **which_ones** (:obj:`str`, *Optional*): {'final': Results of the final iteration, 'blended': Averaged results
            for all iterations, 'all': Saves skims for both the final iteration and the blended ones}.
            Default is 'final'

            **format** (:obj:`str`, *Optional*): File format ('aem' or 'omx'). Default is 'omx'

            **project** (:obj:`Project`, *Optional*): Project we want to save the results to.
            Defaults to the active project
        """
        mat_format = format.lower()
        if mat_format not in ["omx", "aem"]:
            raise ValueError("Matrix needs to be either OMX or native AequilibraE")
            raise ImportError("OpenMatrix is not available on your system")

        if not project:
            project = self.project or get_active_project()

        mats = project.matrices

        for cls in self.classes:
            file_name = f"{matrix_name}_{cls._id}.{mat_format}"

            export_name = path.join(mats.fldr, file_name)

            if path.isfile(export_name):
                raise FileExistsError(f"{file_name} already exists. Choose a different name or matrix format")

            if mats.check_exists(matrix_name):
                raise FileExistsError(f"{matrix_name} already exists. Choose a different name")

            avg_skims = cls.results.skims  # type: AequilibraeMatrix

            # The ones for the last iteration are here
            last_skims = cls._aon_results.skims  # type: AequilibraeMatrix

            names = []
            if which_ones in ["final", "all"]:
                for core in last_skims.names:
                    names.append(f"{core}_final")

            if which_ones in ["blended", "all"]:
                for core in avg_skims.names:
                    names.append(f"{core}_blended")

            if not names:
                continue
            # Assembling a single final skim file can be done like this
            # We will want only the time for the last iteration and the distance averaged out for all iterations
            working_name = export_name if mat_format == "aem" else AequilibraeMatrix().random_name()

            kwargs = {
                "file_name": working_name,
                "zones": self.classes[0].graph.centroids.shape[0],
                "matrix_names": names,
                "memory_only": False,
            }

            # Create the matrix to manipulate
            out_skims = AequilibraeMatrix()
            out_skims.create_empty(**kwargs)

            out_skims.index[:] = self.classes[0].graph.centroids[:]
            out_skims.description = f"Assignment skim from procedure ID {self.procedure_id}. Class name {cls._id}"

            if which_ones in ["final", "all"]:
                for core in last_skims.names:
                    out_skims.matrix[f"{core}_final"][:, :] = last_skims.matrix[core][:, :]

            if which_ones in ["blended", "all"]:
                for core in avg_skims.names:
                    out_skims.matrix[f"{core}_blended"][:, :] = avg_skims.matrix[core][:, :]

            out_skims.matrices.flush()  # Make sure that all data went to the disk

            # If it were supposed to be an OMX, we export to one
            if mat_format == "omx":
                out_skims.export(export_name)

            out_skims.description = f"Skimming for assignment procedure. Class {cls._id}"
            # Now we create the appropriate record

            record = mats.new_record(f"{matrix_name}_{cls._id}", file_name)
            record.procedure_id = self.procedure_id
            record.timestamp = self.procedure_date
            record.procedure = "Traffic Assignment"
            record.description = out_skims.description
            record.save()

    def select_link_flows(self) -> Dict[str, pd.DataFrame]:
        """
        Returns a dataframe of the select link flows for each class
        """
        class_flows = []  # stores the df for each class
        for cls in self.classes:
            # Save OD_matrices
            if cls._selected_links is None:
                continue
            # Create Values table
            df = cls.results.get_sl_results()
            # Remap the dataframe names to add the prefix classname for each class
            cls_cols = {x: cls._id + "_" + x if (x != "index") else "link_id" for x in df.columns}
            df.rename(columns=cls_cols, inplace=True)
            class_flows.append(df)
        return pd.concat(class_flows, axis=1).rename_axis("link_id")

    def save_select_link_flows(self, table_name: str, project=None) -> None:
        """
        Saves the select link link flows for all classes into the results database. Additionally, it exports
        the OD matrices into OMX format.

        :Arguments:
            **table_name** (:obj:`str`): Name of the table being inserted to. Note the traffic class

            **project** (:obj:`Project`, *Optional*): Project we want to save the results to.
            Defaults to the active project
        """

        if not project:
            project = self.project or get_active_project()
        df = self.select_link_flows()
        conn = sqlite3.connect(path.join(project.project_base_path, "results_database.sqlite"))
        df.to_sql(table_name, conn)
        conn.close()
        # Create description table

        self.description = f"Select link analysis from {self.procedure_id}"
        conn = project.connect()
        report = {}
        data = [
            table_name,
            "select link",
            self.procedure_id,
            str(report),
            self.procedure_date,
            self.description,
        ]
        conn.execute(
            """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                            description) Values(?,?,?,?,?,?)""",
            data,
        )
        conn.commit()
        conn.close()

    def save_select_link_matrices(self, file_name: str) -> None:
        """
        Saves the Select Link matrices for each TrafficClass in the current TrafficAssignment class
        """

        for cls in self.classes:
            # Save OD_matrices
            if cls._selected_links is None:
                continue
            cls.results.select_link_od.export(str(Path(file_name).with_suffix(".omx")))

    def save_select_link_results(self, name: str) -> None:
        """
        Saves both the Select Link matrices and flow results at the same time, using the same name.

        .. note::
            Note the Select Link matrices will have _SL_matrices.omx appended to the end for ease of identification.
            e.g. save_select_link_results("Car") will result in the following names for the flows and matrices:
            Select Link Flows: inserts the select link flows for each class into the database with the table name:
            Car
            Select Link Matrices (only exports to OMX format):
            Car.omx

        :Arguments:
            **name** (:obj:`str`): name of the matrices
        """
        self.save_select_link_flows(name)
        self.save_select_link_matrices(name)


class TransitAssignment(AssignmentBase):
    all_algorithms = ["optimal-strategies", "os"]

    def __init__(self, *args, project=None, **kwargs):
        super().__init__(*args, **kwargs)

    def set_algorithm(self, algorithm: str):
        """
        Chooses the assignment algorithm. Currently only 'optimal-strategies' is available.

        'os' is also accepted as an alternative to 'optimal-strategies'

        :Arguments:
            **algorithm** (:obj:`str`): Algorithm to be used
        """
        algo_dict = {i: i for i in self.all_algorithms}
        algo_dict["os"] = "optimal-strategies"
        algo = algo_dict.get(algorithm.lower())

        if algo is None:
            raise AttributeError(f"Assignment algorithm not available. Choose from: {','.join(self.all_algorithms)}")

        self.algorithm = algo
        self._config["Algorithm"] = algo
        self.assignment = OptimalStrategies(self)

    def set_cores(self, cores: int) -> None:
        """Allows one to set the number of cores to be used AFTER transit classes have been added

        Inherited from :obj:`AssignmentResultsBase`

        :Arguments:
            **cores** (:obj:`int`): Number of CPU cores to use
        """
        if not self.classes:
            raise RuntimeError("You need load transit classes before overwriting the number of cores")

        self.cores = set_cores(cores)
        for c in self.classes:
            c.results.set_cores(self.cores)

    def info(self) -> dict:
        """Returns information for the transit assignment procedure

        Dictionary contains keys  'Algorithm', 'Classes', 'Computer name', 'Procedure ID'.

        The classes key is also a dictionary with all the user classes per transit class and their respective
        matrix totals

        :Returns:
            **info** (:obj:`dict`): Dictionary with summary information
        """

        classes = {}

        for cls in self.classes:
            uclass = {}

            if len(cls.matrix.view_names) == 1:
                uclass["matrix_totals"] = {nm: np.sum(cls.matrix.matrix_view[:, :]) for nm in cls.matrix.view_names}
            else:
                uclass["matrix_totals"] = {
                    nm: np.sum(cls.matrix.matrix_view[:, :, i]) for i, nm in enumerate(cls.matrix.view_names)
                }
            uclass["network mode"] = cls.graph.mode

            classes[cls._id] = uclass

        info = {
            "Algorithm": self.algorithm,
            "Classes": classes,
            "Computer name": socket.gethostname(),
            "Procedure ID": self.procedure_id,
        }
        return info

    def log_specification(self):
        self.logger.info("Transit Class specification")
        for cls in self.classes:
            self.logger.info(str(cls.info))

        self.logger.info("Transit Assignment specification")
        self.logger.info(self._config)

    def save_results(self, table_name: str, keep_zero_flows=True, project=None) -> None:
        """Saves the assignment results to results_database.sqlite

        Method fails if table exists

        :Arguments:
            **table_name** (:obj:`str`): Name of the table to hold this assignment result

            **keep_zero_flows** (:obj:`bool`): Whether we should keep records for zero flows. Defaults to ``True``

            **project** (:obj:`Project`, *Optional*): Project we want to save the results to.
            Defaults to the active project
        """

        df = self.results()
        if not keep_zero_flows:
            df = df[df.volume > 0]

        if not project:
            project = project or get_active_project()
        conn = sqlite3.connect(path.join(project.project_base_path, "results_database.sqlite"))
        df.to_sql(table_name, conn)
        conn.close()

        conn = database_connection("transit", project.project_base_path)
        report = {"setup": self.info()}
        data = [table_name, "transit assignment", self.procedure_id, str(report), self.procedure_date, self.description]
        conn.execute(
            """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                            description) Values(?,?,?,?,?,?)""",
            data,
        )
        conn.commit()
        conn.close()

    def results(self) -> pd.DataFrame:
        """Prepares the assignment results as a Pandas DataFrame

        :Returns:
            **DataFrame** (:obj:`pd.DataFrame`): Pandas DataFrame with all the assignment results indexed on *link_id*
        """
        assig_results = [
            pd.DataFrame(cls.results.get_load_results()).rename(columns={"volume": cls._id + "_volume"})
            for cls in self.classes
        ]

        return pd.concat(assig_results, axis=1)

    def set_time_field(self, time_field: str) -> None:
        """
        Sets the graph field that contains free flow travel time -> e.g. 'trav_time'

        :Arguments:
            **time_field** (:obj:`str`): Field name
        """
        super().set_time_field(time_field)
        self._config["Time field"] = time_field

    def set_frequency_field(self, frequency_field: str) -> None:
        """
        Sets the graph field that contains the frequency -> e.g. 'freq'

        :Arguments:
            **frequency_field** (:obj:`str`): Field name
        """
        self._check_field(frequency_field)
        self._config["Frequency field"] = frequency_field



from io import TextIOWrapper
import numpy as np
import csv
import copy
from numpy.lib.recfunctions import append_fields


def parse_csv(file_name: str, column_order=[]):  # noqa B006
    tot = []
    if isinstance(file_name, str):
        csvfile = open(file_name, encoding="utf-8-sig")
    else:
        csvfile = TextIOWrapper(file_name, encoding="utf-8-sig")

    contents = csv.reader(csvfile, delimiter=",", quotechar='"')
    numcols = 0
    for row in contents:
        if not len("".join(row).strip()):
            continue
        broken = [x.encode("ascii", errors="ignore").decode().strip() for x in row]

        if not numcols:
            numcols = len(broken)
        else:
            if numcols < len(broken):
                broken.extend([""] * (numcols - len(broken)))

        tot.append(broken)
    titles = tot.pop(0)
    csvfile.close()
    if tot:
        data = np.core.records.fromrecords(tot, names=[x.lower() for x in titles])
    else:
        return empty()

    missing_cols_names = [x for x in column_order.keys() if x not in data.dtype.names]
    for col in missing_cols_names:
        data = append_fields(data, col, np.array([""] * len(tot)))

    if column_order:
        col_names = [x for x in column_order.keys() if x in data.dtype.names]
        data = data[col_names]

        # Define sizes for the string variables
        column_order = copy.deepcopy(column_order)
        for c in col_names:
            if column_order[c] is str:
                column_order[c] = object
            else:
                if data[c].dtype.char.upper() in ["U", "S"]:
                    data[c][data[c] == ""] = "0"

        new_data_dt = [(f, column_order[f]) for f in col_names]

        if int(data.shape.__len__()) > 0:
            # handle the case of int data given as a float string
            for j, (_, dtype) in enumerate(new_data_dt):
                if dtype is int:
                    for item in data:
                        item[j] = item[j].split(".")[0]

            return np.array(data, dtype=new_data_dt)
        else:
            return data
    else:
        return data


class empty:
    shape = [0]



from collections import OrderedDict

column_order = {
    "agency.txt": OrderedDict(
        [
            ("agency_id", str),
            ("agency_name", str),
            ("agency_url", str),
            ("agency_timezone", str),
            ("agency_lang", str),
            ("agency_phone", str),
            ("agency_fare_url", str),
            ("agency_email", str),
        ]
    ),
    "routes.txt": OrderedDict(
        [
            ("route_id", str),
            ("route_short_name", str),
            ("route_long_name", str),
            ("route_desc", str),
            ("route_type", int),
            # ("route_url", str),
            # ("route_color", str),
            # ("route_text_color", str),
            # ("route_sort_order", int),
            # ("agency_id", str),
        ]
    ),
    "trips.txt": OrderedDict(
        [
            ("route_id", str),
            ("service_id", str),
            ("trip_id", str),
            # ("trip_headsign", str),
            # ("trip_short_name", str),
            # ("block_id", str),
            ("shape_id", str),
            ("direction_id", int),
            # ("wheelchair_accessible", int),
            # ("bikes_allowed", int),
        ]
    ),
    "stop_times.txt": OrderedDict(
        [
            ("trip_id", str),
            ("arrival_time", str),
            ("departure_time", str),
            ("stop_id", str),
            ("stop_sequence", int),
            # ("stop_headsign", str),
            # ("pickup_type", int),
            # ("shape_dist_traveled", float),
            # ("timepoint", int),
        ]
    ),
    "calendar.txt": OrderedDict(
        [
            ("service_id", str),
            ("monday", int),
            ("tuesday", int),
            ("wednesday", int),
            ("thursday", int),
            ("friday", int),
            ("saturday", int),
            ("sunday", int),
            ("start_date", str),
            ("end_date", str),
        ]
    ),
    "calendar_dates.txt": OrderedDict([("service_id", str), ("date", str), ("exception_type", int)]),
    "fare_attributes.txt": OrderedDict(
        [
            ("fare_id", str),
            ("price", float),
            ("currency_type", str),
            ("payment_method", int),
            ("transfers", int),
            # ("agency_id", str),
            ("transfer_duration", float),
        ]
    ),
    "fare_rules.txt": OrderedDict(
        [("fare_id", str), ("route_id", str), ("origin_id", str), ("destination_id", str), ("contains_id", str)]
    ),
    "frequencies.txt": OrderedDict(
        [
            ("trip_id", str),
            ("start_time", str),
            ("end_time", str),
            ("headway_secs", str),
            # ("exact_times", int)
        ]
    ),
    "transfers.txt": OrderedDict(
        [("from_stop_id", str), ("to_stop_id", str), ("transfer_type", int), ("min_transfer_time", int)]
    ),
    "feed_info.txt": OrderedDict(
        [
            ("feed_publisher_name", str),
            ("feed_publisher_url", str),
            ("feed_lang", str),
            ("feed_start_date", str),
            ("feed_end_date", str),
            ("feed_version", str),
        ]
    ),
    "stops.txt": OrderedDict(
        [
            ("stop_id", str),
            # ("stop_code", str),
            ("stop_name", str),
            ("stop_desc", str),
            ("stop_lat", float),
            ("stop_lon", float),
            ("stop_street", str),
            ("zone_id", str),
            # ("stop_url", str),
            # ("location_type", int),
            ("parent_station", str),
            # ("stop_timezone", str),
            # ("wheelchair_boarding", int),
        ]
    ),
    "shapes.txt": OrderedDict(
        [
            ("shape_id", str),
            ("shape_pt_lat", float),
            ("shape_pt_lon", float),
            ("shape_pt_sequence", int),
            # ("shape_dist_traveled", float),
        ]
    ),
}



from sqlite3 import Connection
from typing import Optional

from shapely.geometry import LineString, Polygon
from aequilibrae.utils.db_utils import commit_and_close

INFINITE_CAPACITY = 99999


def connector_creation(
    zone_id: int,
    mode_id: str,
    network,
    link_types="",
    connectors=1,
    conn_: Optional[Connection] = None,
    delimiting_area: Polygon = None,
):
    if len(mode_id) > 1:
        raise Exception("We can only add centroid connectors for one mode at a time")

    with conn_ or commit_and_close(network.project.path_to_file) as conn:
        logger = network.project.logger
        if sum(conn.execute("select count(*) from nodes where node_id=?", [zone_id]).fetchone()) == 0:
            logger.warning("This centroid does not exist. Please create it first")
            return

        sql = "select count(*) from links where a_node=? and instr(modes,?) > 0"
        if conn.execute(sql, [zone_id, mode_id]).fetchone()[0] > 0:
            logger.warning("Mode is already connected")
            return

    proj_nodes = network.nodes.data

    centroid = proj_nodes.query("node_id == @zone_id")  # type: gpd.GeoDataFrame
    centroid = centroid.rename(columns={"node_id": "zone_id"})[["zone_id", "geometry"]]

    nodes = proj_nodes.query("is_centroid != 1 and modes.str.contains(@mode_id)", engine="python")

    if len(link_types) > 0:
        nodes = nodes[nodes.link_types.str.contains("|".join(list(link_types)))]

    if delimiting_area is not None:
        nodes = nodes[nodes.geometry.within(delimiting_area)]

    if nodes.empty:
        zone_id = centroid["zone_id"].values[0]
        logger.warning(f"No nodes found for centroid {zone_id} (mode {mode_id} and link types {link_types})")
        return

    joined = nodes[["node_id", "geometry"]].sjoin_nearest(centroid, distance_col="distance_connector")
    joined = joined.nsmallest(connectors, "distance_connector")

    centr_geo = centroid.geometry.values[0]
    links = network.links
    for _, rec in joined.iterrows():
        link = links.new()
        link.geometry = LineString([centr_geo, rec.geometry])
        link.modes = mode_id
        link.direction = 0
        link.link_type = "centroid_connector"
        link.name = f"centroid connector zone {zone_id}"
        link.capacity_ab = INFINITE_CAPACITY
        link.capacity_ba = INFINITE_CAPACITY
        link.save()
    if not joined.empty:
        logger.warning(f"{joined.shape[0]} new centroid connectors for mode {mode_id} added for centroid {zone_id}")



import pandas as pd


class BasicPTElement:
    def from_row(self, data: pd.Series):
        for key, value in data.iteritems():
            if key not in self.__dict__.keys():
                raise KeyError(f"{key} Field does not exist")
            self.__dict__[key] = value
        return self



import importlib.util as iutil
import platform
from os.path import join

import numpy as np
from Cython.Distutils import build_ext
from Cython.Build import cythonize
from setuptools import Extension
from setuptools import setup, find_packages
from setuptools.discovery import FlatLayoutPackageFinder

# When updating the version, one must also update the docs/source/_static/switcher.json file
version = 1.1
minor_version = 3

release_version = f"{version}.{minor_version}"

include_dirs = [np.get_include()]
libraries = []
library_dirs = []
if iutil.find_spec("pyarrow") is not None:
    import pyarrow as pa

    pa.create_library_symlinks()
    include_dirs.append(pa.get_include())
    libraries.extend(pa.get_libraries())
    library_dirs.extend(pa.get_library_dirs())

is_win = "WINDOWS" in platform.platform().upper()
is_mac = any(e in platform.platform().upper() for e in ["MACOS", "DARWIN"])
prefix = "/" if is_win else "-f"
cpp_std = "/std:c++17" if is_win else "-std=c++17"
compile_args = [cpp_std, f"{prefix}openmp{':llvm' if is_win else ''}"]
compile_args += ["-Wno-unreachable-code"] if is_mac else []
link_args = [f"{prefix}openmp"]

extension_args = {
    "extra_compile_args": compile_args,
    "extra_link_args": link_args,
    "define_macros": [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    "include_dirs": include_dirs,
    "libraries": libraries,
    "library_dirs": library_dirs,
    "language": "c++",
}

ext_mod_aon = Extension("aequilibrae.paths.AoN", [join("aequilibrae", "paths", "cython", "AoN.pyx")], **extension_args)

ext_mod_ipf = Extension(
    "aequilibrae.distribution.ipf_core",
    [join("aequilibrae", "distribution", "cython", "ipf_core.pyx")],
    **extension_args,
)

ext_mod_put = Extension(
    "aequilibrae.paths.public_transport",
    [join("aequilibrae", "paths", "cython", "public_transport.pyx")],
    **extension_args,
)

ext_mod_rc = Extension(
    "aequilibrae.paths.cython.route_choice_set",
    [join("aequilibrae", "paths", "cython", "route_choice_set.pyx")],
    **extension_args,
)

ext_mod_rc_ll_results = Extension(
    "aequilibrae.paths.cython.route_choice_link_loading_results",
    [join("aequilibrae", "paths", "cython", "route_choice_link_loading_results.pyx")],
    **extension_args,
)

ext_mod_rc_set_results = Extension(
    "aequilibrae.paths.cython.route_choice_set_results",
    [join("aequilibrae", "paths", "cython", "route_choice_set_results.pyx")],
    **extension_args,
)

ext_mod_graph_building = Extension(
    "aequilibrae.paths.graph_building",
    [join("aequilibrae", "paths", "cython", "graph_building.pyx")],
    **extension_args,
)

ext_mod_sparse_matrix = Extension(
    "aequilibrae.matrix.sparse_matrix",
    [join("aequilibrae", "matrix", "sparse_matrix.pyx")],
    **extension_args,
)

ext_mod_coo_demand = Extension(
    "aequilibrae.matrix.coo_demand",
    [join("aequilibrae", "matrix", "coo_demand.pyx")],
    **extension_args,
)

with open("requirements.txt", "r") as fl:
    install_requirements = [x.strip() for x in fl.readlines()]

pkgs = find_packages(exclude=FlatLayoutPackageFinder.DEFAULT_EXCLUDE)

pkg_data = {
    "aequilibrae.reference_files": ["spatialite.sqlite", "nauru.zip", "sioux_falls.zip", "coquimbo.zip"],
    "aequilibrae.paths": ["cython/*.pxi", "cython/*.pyx", "cython/*.pxd"],
    "aequilibrae.distribution": ["cython/*.pyx"],
    "aequilibrae.matrix": ["*.pyx", "*.pxd"],
    "aequilibrae": ["./parameters.yml", "../requirements.txt"],
    "aequilibrae.project": [
        "database_specification/network/tables/*.*",
        "database_specification/network/triggers/*.*",
        "database_specification/transit/tables/*.*",
        "database_specification/transit/triggers/*.*",
    ],
}

with open("README.md", "r") as fh:
    long_description = fh.read()

if __name__ == "__main__":
    setup(
        name="aequilibrae",
        version=release_version,  # noqa: F821
        install_requires=install_requirements,
        packages=pkgs,
        package_dir={"": "."},
        package_data=pkg_data,
        zip_safe=False,
        description="A package for transportation modeling",
        long_description=long_description,
        author="Pedro Camargo",
        author_email="c@margo.co",
        url="https://github.com/AequilibraE/aequilibrae",
        license="See LICENSE.TXT",
        license_files=("LICENSE.TXT",),
        classifiers=[
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
        ],
        cmdclass={"build_ext": build_ext},
        ext_modules=cythonize(
            [
                ext_mod_aon,
                ext_mod_ipf,
                ext_mod_put,
                ext_mod_rc,
                ext_mod_coo_demand,
                ext_mod_rc_ll_results,
                ext_mod_rc_set_results,
                ext_mod_graph_building,
                ext_mod_sparse_matrix,
            ],
            compiler_directives={"language_level": "3str"},
        ),
    )



import numpy as np

from aequilibrae.paths.multi_threaded_paths import MultiThreadedPaths


class MultiThreadedNetworkSkimming(MultiThreadedPaths):
    def __init__(self):
        MultiThreadedPaths.__init__(self)

        # holds the skims for all nodes in the network (during path finding)
        self.temporary_skims = np.array([], np.int64)

    # In case we want to do by hand, we can prepare each method individually
    def prepare(self, graph, cores, nodes, num_skims):
        self.prepare_(graph, cores, nodes)

        ftype = graph.default_types("float")
        self.temporary_skims = np.zeros((cores, nodes, num_skims), dtype=ftype)



# cython: language_level=3str
from aequilibrae.paths.graph import Graph
from aequilibrae.paths.cython.route_choice_types cimport LinkSet_t, minstd_rand, shuffle
from aequilibrae.matrix.coo_demand cimport GeneralisedCOODemand

from cython.operator cimport dereference as d
from cython.parallel cimport parallel, prange, threadid
from libc.limits cimport UINT_MAX
from libc.string cimport memcpy
from libcpp cimport nullptr
from libcpp.algorithm cimport reverse, copy
from libcpp.unordered_set cimport unordered_set
from libcpp.utility cimport pair
from libcpp.vector cimport vector
from libcpp cimport bool
from openmp cimport omp_get_max_threads

from libcpp.memory cimport shared_ptr

from typing import Tuple

import numpy as np
import pyarrow as pa
import pandas as pd


"""This module aims to implemented the BFS-LE algorithm as described in Rieser-Schüssler, Balmer, and Axhausen, 'Route
Choice Sets for Very High-Resolution Data'.  https://doi.org/10.1080/18128602.2012.671383

A rough overview of the algorithm is as follows.  1. Prepare the initial graph, this is depth 0 with no links removed.
    2. Find a short path, P. If P is not empty add P to the path set.  3. For all links p in P, remove p from E,
    compounding with the previously removed links.  4. De-duplicate the sub-graphs, we only care about unique
    sub-graphs.  5. Go to 2.

Details: The general idea of the algorithm is pretty simple, as is the implementation. The caveats here is that there is
a lot of cpp interop and memory management. A description of the purpose of variables is in order:

route_set: See route_choice.pxd for full type signature. It's an unordered set (hash set) of pointers to vectors of link
IDs. It uses a custom hashing function and comparator. The hashing function is defined in a string that in inlined
directly into the output ccp. This is done allow declaring of the `()` operator, which is required and AFAIK not
possible in Cython. The hash is designed to dereference then hash order dependent vectors. One isn't provided by
stdlib. The comparator simply dereferences the pointer and uses the vector comparator. It's designed to store the
outputted paths. Heap allocated (needs to be returned).

removed_links: See route_choice.pxd for full type signature. It's an unordered set of pointers to unordered sets of link
IDs. Similarly to `route_set` is uses a custom hash function and comparator. This hash function is designed to be order
independent and should only use commutative operations. The comparator is the same. It's designed to store all of the
removed link sets we've seen before. This allows us to detected duplicated graphs.

rng: A custom imported version of std::linear_congruential_engine. libcpp doesn't provide one so we do. It should be
significantly faster than the std::mersenne_twister_engine without sacrificing much. We don't need amazing RNG, just ok
and fast. This is only used to shuffle the queue.

queue, next_queue: These are vectors of pointers to sets of removed links. We never need to push to the front of these
so a vector is best. We maintain two queues, one that we are currently iterating over, and one that we can add to,
building up with all the newly removed link sets. These two are swapped at the end of an iteration, next_queue is then
cleared. These store sets of removed links.

banned, next_banned: `banned` is the iterator variable for `queue`. `banned` is copied into `next_banned` where another
link can be added without mutating `banned`. If we've already seen this set of removed links `next_banned` is
immediately deallocated. Otherwise it's placed into `next_queue`.

vec: `vec` is a scratch variable to store pointers to new vectors, or rather, paths while we are building them. Each
time a path is found a new one is allocated, built, and stored in the route_set.

p, connector: Scratch variables for iteration.

Optimisations: As described in the paper, both optimisations have been implemented. The path finding operates on the
compressed graph and the queue is shuffled if its possible to fill the route set that iteration. The route set may not
be filled due to duplicate paths but we can't know that in advance so we shuffle anyway.

Any further optimisations should focus on the path finding, from benchmarks it dominates the run time (~98%). Since huge
routes aren't required small-ish things like the memcpy and banned link set copy aren't high priority.

"""

# It would really be nice if these were modules. The 'include' syntax is long deprecated and adds a lot to compilation
# times
include 'basic_path_finding.pyx'


@cython.embedsignature(True)
cdef class RouteChoiceSet:
    """
    Route choice implemented via breadth first search with link removal (BFS-LE) as described in Rieser-Schüssler,
    Balmer, and Axhausen, 'Route Choice Sets for Very High-Resolution Data'
    """

    def __init__(self, graph: Graph):
        """Python level init, may be called multiple times, for things that can't be done in __cinit__."""
        # self.heuristic = HEURISTIC_MAP[self.res._heuristic]
        self.cost_view = graph.compact_cost
        self.graph_fs_view = graph.compact_fs
        self.b_nodes_view = graph.compact_graph.b_node.values
        self.nodes_to_indices_view = graph.compact_nodes_to_indices

        # tmp = graph.lonlat_index.loc[graph.compact_all_nodes]
        # self.lat_view = tmp.lat.values
        # self.lon_view = tmp.lon.values
        self.a_star = False

        self.ids_graph_view = graph.compact_graph.id.values

        # We explicitly don't want the links that have been removed from the graph
        self.graph_compressed_id_view = graph.graph.__compressed_id__.values
        self.num_nodes = graph.compact_num_nodes
        self.num_links = graph.compact_num_links
        self.zones = graph.num_zones
        self.block_flows_through_centroids = graph.block_centroid_flows

        self.mapping_idx, self.mapping_data, _ = graph.create_compressed_link_network_mapping()

        self.results = None
        self.ll_results = None

    @cython.embedsignature(True)
    def run(self, origin: int, destination: int, shape: Tuple[int, int], demand: float = 0.0, *args, **kwargs):
        """Compute the a route set for a single OD pair.

        Often the returned list's length is ``max_routes``, however, it may be limited by ``max_depth`` or if all
        unique possible paths have been found then a smaller set will be returned.

        Additional arguments are forwarded to ``RouteChoiceSet.batched``.

        :Arguments:
            **origin** (:obj:`int`): Origin node ID. Must be present within compact graph. Recommended to choose a
                centroid.
            **destination** (:obj:`int`): Destination node ID. Must be present within compact graph. Recommended to
                choose a centroid.
            **demand** (:obj:`double`): Demand for this single OD pair.

        :Returns: **route set** (:obj:`list[tuple[int, ...]]): Returns a list of unique variable length tuples of
            link IDs. Represents paths from ``origin`` to ``destination``.
        """
        df = pd.DataFrame({
            "origin id": [origin],
            "destination id": [destination],
            "demand": [demand]
        }).set_index(["origin id", "destination id"])
        demand_coo = GeneralisedCOODemand("origin id", "destination id", np.asarray(self.nodes_to_indices_view), shape)
        demand_coo.add_df(df)

        self.batched(demand_coo, {}, *args, **kwargs)
        where = kwargs.get("where", None)
        if where is not None:
            schema = self.psl_schema if kwargs.get("path_size_logit", False) else self.schema
            results = pa.dataset.dataset(
                where, format="parquet", partitioning=pa.dataset.HivePartitioning(schema)
            ).to_table()
        else:
            results = self.get_results()
        return [tuple(x) for x in results.column("route set").to_pylist()]

    # Bounds checking doesn't really need to be disabled here but the warning is annoying
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.initializedcheck(False)
    def batched(
            self,
            demand: GeneralisedCOODemand,
            select_links: Dict[str, FrozenSet[FrozenSet[int]]] = None,
            sl_link_loading: bool = True,
            max_routes: int = 0,
            max_depth: int = 0,
            max_misses: int = 100,
            seed: int = 0,
            cores: int = 0,
            a_star: bool = True,
            bfsle: bool = True,
            penalty: float = 1.0,
            where: Optional[str] = None,
            store_results: bool = True,
            path_size_logit: bool = False,
            beta: float = 1.0,
            cutoff_prob: float = 0.0,
    ):
        """Compute the a route set for a list of OD pairs.

        Often the returned list for each OD pair's length is ``max_routes``, however, it may be limited by ``max_depth``
        or if all unique possible paths have been found then a smaller set will be returned.

        :Arguments:
            **ods** (:obj:`list[tuple[int, int]]`): List of OD pairs ``(origin, destination)``. Origin and destination
                node ID must be present within compact graph. Recommended to choose a centroids.
            **max_routes** (:obj:`int`): Maximum size of the generated route set. Must be non-negative. Default of
                ``0`` for unlimited.
            **max_depth** (:obj:`int`): Maximum depth BFSLE can explore, or maximum number of iterations for link
                penalisation. Must be non-negative. Default of ``0`` for unlimited.
            **max_misses** (:obj:`int`): Maximum number of collective duplicate routes found for a single OD pair.
                Terminates if exceeded.
            **seed** (:obj:`int`): Seed used for rng. Must be non-negative. Default of ``0``.
            **cores** (:obj:`int`): Number of cores to use when parallelising over OD pairs. Must be non-negative.
                Default of ``0`` for all available.
            **bfsle** (:obj:`bool`): Whether to use Breadth First Search with Link Removal (BFSLE) over link
                penalisation. Default ``True``.
            **penalty** (:obj:`float`): Penalty to use for Link Penalisation and BFSLE with LP.
            **where** (:obj:`str`): Optional file path to save results to immediately. Will return None.
        """
        cdef:
            long long origin, dest
            size_t i

        if select_links is None:
            select_links = {}

        if max_routes == 0 and max_depth == 0:
            raise ValueError("Either `max_routes` or `max_depth` must be > 0")

        if max_routes < 0 or max_depth < 0:
            raise ValueError("`max_routes`, `max_depth`, and `cores` must be non-negative")

        if path_size_logit and beta < 0:
            raise ValueError("`beta` must be >= 0 for path sized logit model")

        if path_size_logit and not 0.0 <= cutoff_prob <= 1.0:
            raise ValueError("`cutoff_prob` must be 0 <= `cutoff_prob` <= 1 for path sized logit model")

        for origin, dest in demand.df.index:
            if self.nodes_to_indices_view[origin] == -1:
                raise ValueError(f"Origin {origin} is not present within the compact graph")
            if self.nodes_to_indices_view[dest] == -1:
                raise ValueError(f"Destination {dest} is not present within the compact graph")

        cdef:
            long long origin_index, dest_index
            unsigned int c_max_routes = max_routes
            unsigned int c_max_depth = max_depth
            unsigned int c_max_misses = max_misses
            unsigned int c_seed = seed
            unsigned int c_cores = cores if cores > 0 else omp_get_max_threads()

            # Scale cutoff prob from [0, 1] -> [0.5, 1]. Values below 0.5 produce negative inverse binary logit values.
            double scaled_cutoff_prob = (1.0 - cutoff_prob) * 0.5 + 0.5

            # A* (and Dijkstra's) require memory views, so we must allocate here and take slices. Python can handle this
            # memory
            double [:, :] cost_matrix = np.empty((c_cores, self.cost_view.shape[0]), dtype=float)
            long long [:, :] predecessors_matrix = np.empty((c_cores, self.num_nodes + 1), dtype=np.int64)
            long long [:, :] conn_matrix = np.empty((c_cores, self.num_nodes + 1), dtype=np.int64)
            long long [:, :] b_nodes_matrix = np.broadcast_to(
                self.b_nodes_view,
                (c_cores, self.b_nodes_view.shape[0])
            ).copy()

            # This matrix is never read from, it exists to allow using the Dijkstra's method without changing the
            # interface.
            long long [:, :] _reached_first_matrix

        # self.a_star = a_star

        pa.set_io_thread_count(c_cores)

        if self.a_star:
            _reached_first_matrix = np.zeros((c_cores, 1), dtype=np.int64)  # Dummy array to allow slicing
        else:
            _reached_first_matrix = np.zeros((c_cores, self.num_nodes + 1), dtype=np.int64)

        cdef:
            RouteSet_t *route_set
            shared_ptr[vector[double]] prob_vec
            int thread_id

        demand._initalise_col_names()
        self.ll_results = LinkLoadingResults(demand, select_links, self.num_links, sl_link_loading, c_cores)

        for _, grouped_demand_df in (demand.batches() if where is not None else ((None, None),)):
            demand._initalise_c_data(grouped_demand_df)

            self.results = RouteChoiceSetResults(
                demand,
                scaled_cutoff_prob,
                beta,
                self.num_links,
                self.cost_view,
                self.mapping_idx,
                self.mapping_data,
                store_results=store_results,
                perform_assignment=path_size_logit,
            )

            with nogil, parallel(num_threads=c_cores):
                route_set = new RouteSet_t()
                thread_id = threadid()
                for i in prange(demand.ods.size()):
                    origin_index = self.nodes_to_indices_view[demand.ods[i].first]
                    dest_index = self.nodes_to_indices_view[demand.ods[i].second]

                    if origin_index == dest_index:
                        continue

                    if self.block_flows_through_centroids:
                        blocking_centroid_flows(
                            0,  # Always blocking
                            origin_index,
                            self.zones,
                            self.graph_fs_view,
                            b_nodes_matrix[thread_id],
                            self.b_nodes_view,
                        )

                    if bfsle:
                        RouteChoiceSet.bfsle(
                            self,
                            d(route_set),
                            origin_index,
                            dest_index,
                            c_max_routes,
                            c_max_depth,
                            c_max_misses,
                            cost_matrix[thread_id],
                            predecessors_matrix[thread_id],
                            conn_matrix[thread_id],
                            b_nodes_matrix[thread_id],
                            _reached_first_matrix[thread_id],
                            penalty,
                            c_seed,
                        )
                    else:
                        RouteChoiceSet.link_penalisation(
                            self,
                            d(route_set),
                            origin_index,
                            dest_index,
                            c_max_routes,
                            c_max_depth,
                            c_max_misses,
                            cost_matrix[thread_id],
                            predecessors_matrix[thread_id],
                            conn_matrix[thread_id],
                            b_nodes_matrix[thread_id],
                            _reached_first_matrix[thread_id],
                            penalty,
                            c_seed,
                        )

                    # Here we transform the set of raw pointers to routes (vectors) into a vector of unique points to
                    # routes. This is done to simplify memory management later on.
                    route_vec = self.results.get_route_vec(i)
                    RouteChoiceSetResults.route_set_to_route_vec(d(route_vec), d(route_set))
                    # We now drop all references to those raw pointers. The unique pointers now own those vectors.
                    route_set.clear()

                    if path_size_logit:
                        prob_vec = self.results.compute_result(i, d(route_vec), thread_id)
                        self.ll_results.link_load_single_route_set(i, d(route_vec), d(prob_vec), thread_id)
                        self.ll_results.sl_link_load_single_route_set(
                            i, d(route_vec),
                            d(prob_vec),
                            origin_index,
                            dest_index,
                            thread_id
                        )

                    if self.block_flows_through_centroids:
                        blocking_centroid_flows(
                            1,  # Always unblocking
                            origin_index,
                            self.zones,
                            self.graph_fs_view,
                            b_nodes_matrix[thread_id],
                            self.b_nodes_view,
                        )

                del route_set

            if store_results:
                self.get_results()
                if where is not None:
                    self.results.write(where)

        if path_size_logit:
            self.ll_results.reduce_link_loading()
            self.ll_results.reduce_sl_link_loading()
            self.ll_results.reduce_sl_od_matrix()

            self.get_link_loading(cores=c_cores)
            self.get_sl_link_loading(cores=c_cores)
            self.get_sl_od_matrices()

    @cython.initializedcheck(False)
    cdef void path_find(
        RouteChoiceSet self,
        long origin_index,
        long dest_index,
        double [:] thread_cost,
        long long [:] thread_predecessors,
        long long [:] thread_conn,
        long long [:] thread_b_nodes,
        long long [:] _thread_reached_first
    ) noexcept nogil:
        """Small wrapper around path finding, thread locals should be passes as arguments."""
        if self.a_star:
            path_finding_a_star(
                origin_index,
                dest_index,
                thread_cost,
                thread_b_nodes,
                self.graph_fs_view,
                self.nodes_to_indices_view,
                self.lat_view,
                self.lon_view,
                thread_predecessors,
                self.ids_graph_view,
                thread_conn,
                EQUIRECTANGULAR  # FIXME: enum import failing due to redefinition
            )
        else:
            path_finding(
                origin_index,
                dest_index,
                thread_cost,
                thread_b_nodes,
                self.graph_fs_view,
                thread_predecessors,
                self.ids_graph_view,
                thread_conn,
                _thread_reached_first
            )

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.initializedcheck(False)
    cdef void bfsle(
        RouteChoiceSet self,
        RouteSet_t &route_set,
        long origin_index,
        long dest_index,
        unsigned int max_routes,
        unsigned int max_depth,
        unsigned int max_misses,
        double [:] thread_cost,
        long long [:] thread_predecessors,
        long long [:] thread_conn,
        long long [:] thread_b_nodes,
        long long [:] _thread_reached_first,
        double penalty,
        unsigned int seed
    ) noexcept nogil:
        """Main method for route set generation. See top of file for commentary."""
        cdef:
            # Scratch objects
            LinkSet_t removed_links
            minstd_rand rng

            # These objects are juggled to prevent more allocations than necessary
            vector[unordered_set[long long] *] queue
            vector[unordered_set[long long] *] next_queue
            unordered_set[long long] *banned
            unordered_set[long long] *new_banned

            # Local variables, Cython doesn't allow conditional declarations
            vector[long long] *vec
            pair[RouteSet_t.iterator, bool] status
            unsigned int miss_count = 0
            long long p, connector

            # Link penalisation, only used when penalty != 1.0
            bint lp = penalty != 1.0
            vector[double] *penalised_cost = <vector[double] *>nullptr
            vector[double] *next_penalised_cost = <vector[double] *>nullptr

        max_routes = max_routes if max_routes != 0 else UINT_MAX
        max_depth = max_depth if max_depth != 0 else UINT_MAX

        queue.push_back(new unordered_set[long long]())  # Start with no edges banned
        rng.seed(seed)

        if lp:
            # Although we don't need the dynamic ability of vectors here, Cython doesn't have the std::array module.
            penalised_cost = new vector[double](self.cost_view.shape[0])
            next_penalised_cost = new vector[double](self.cost_view.shape[0])
            copy(&self.cost_view[0], &self.cost_view[0] + self.cost_view.shape[0], penalised_cost.begin())
            copy(&self.cost_view[0], &self.cost_view[0] + self.cost_view.shape[0], next_penalised_cost.begin())

        # We'll go at most `max_depth` iterations down, at each depth we maintain a queue of the next set of banned
        # edges to consider
        for depth in range(max_depth):
            if miss_count > max_misses or route_set.size() >= max_routes or queue.size() == 0:
                break

            # If we could potentially fill the route_set after this depth, shuffle the queue
            if queue.size() + route_set.size() >= max_routes:
                shuffle(queue.begin(), queue.end(), rng)

            for banned in queue:
                if lp:
                    # We copy the penalised cost buffer into the thread cost buffer to allow us to apply link
                    # penalisation,
                    copy(penalised_cost.cbegin(), penalised_cost.cend(), &thread_cost[0])
                else:
                    # ...otherwise we just copy directly from the cost view.
                    memcpy(&thread_cost[0], &self.cost_view[0], self.cost_view.shape[0] * sizeof(double))

                for connector in d(banned):
                    thread_cost[connector] = INFINITY

                RouteChoiceSet.path_find(
                    self,
                    origin_index,
                    dest_index,
                    thread_cost,
                    thread_predecessors,
                    thread_conn,
                    thread_b_nodes,
                    _thread_reached_first
                )

                # Mark this set of banned links as seen
                removed_links.insert(banned)

                # If the destination is reachable we must build the path and readd
                if thread_predecessors[dest_index] >= 0:
                    vec = new vector[long long]()
                    # Walk the predecessors tree to find our path, we build it up in a C++ vector because we can't know
                    # how long it'll be
                    p = dest_index
                    while p != origin_index:
                        connector = thread_conn[p]
                        p = thread_predecessors[p]
                        vec.push_back(connector)

                    if lp:
                        # Here we penalise all seen links for the *next* depth. If we penalised on the current depth
                        # then we would introduce a bias for earlier seen paths
                        for connector in d(vec):
                            # *= does not work
                            d(next_penalised_cost)[connector] = penalty * d(next_penalised_cost)[connector]

                    reverse(vec.begin(), vec.end())

                    for connector in d(vec):
                        # This is one area for potential improvement. Here we construct a new set from the old one,
                        # copying all the elements then add a single element. An incremental set hash function could be
                        # of use. However, the since of this set is directly dependent on the current depth and as the
                        # route set size grows so incredibly fast the depth will rarely get high enough for this to
                        # matter. Copy the previously banned links, then for each vector in the path we add one and
                        # push it onto our queue
                        new_banned = new unordered_set[long long](d(banned))
                        new_banned.insert(connector)
                        # If we've already seen this set of removed links before we already know what the path is and
                        # its in our route set
                        if removed_links.find(new_banned) != removed_links.end():
                            del new_banned
                        else:
                            next_queue.push_back(new_banned)

                    # The deduplication of routes occurs here
                    status = route_set.insert(vec)
                    miss_count = miss_count + (not status.second)
                    if miss_count > max_misses or route_set.size() >= max_routes:
                        break

            queue.swap(next_queue)
            next_queue.clear()

            if lp:
                # Update the penalised_cost vector, since next_penalised_cost is always the one updated we just need to
                # bring penalised_cost up to date.
                copy(next_penalised_cost.cbegin(), next_penalised_cost.cend(), penalised_cost.begin())

        # We may have added more banned link sets to the queue then found out we hit the max depth, we should free those
        for banned in queue:
            del banned

        # We should also free all the sets in removed_links, we don't be needing them
        for banned in removed_links:
            del banned

        if lp:
            # If we had enabled link penalisation, we'll need to free those vectors as well
            del penalised_cost
            del next_penalised_cost

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void link_penalisation(
        RouteChoiceSet self,
        RouteSet_t &route_set,
        long origin_index,
        long dest_index,
        unsigned int max_routes,
        unsigned int max_depth,
        unsigned int max_misses,
        double [:] thread_cost,
        long long [:] thread_predecessors,
        long long [:] thread_conn,
        long long [:] thread_b_nodes,
        long long [:] _thread_reached_first,
        double penalty,
        unsigned int seed
    ) noexcept nogil:
        """Link penalisation algorithm for choice set generation."""
        cdef:
            # Scratch objects
            vector[long long] *vec
            long long p, connector
            pair[RouteSet_t.iterator, bool] status
            unsigned int miss_count = 0

        max_routes = max_routes if max_routes != 0 else UINT_MAX
        max_depth = max_depth if max_depth != 0 else UINT_MAX
        memcpy(&thread_cost[0], &self.cost_view[0], self.cost_view.shape[0] * sizeof(double))

        for depth in range(max_depth):
            if route_set.size() >= max_routes:
                break

            RouteChoiceSet.path_find(
                self,
                origin_index,
                dest_index,
                thread_cost,
                thread_predecessors,
                thread_conn,
                thread_b_nodes,
                _thread_reached_first
            )

            if thread_predecessors[dest_index] >= 0:
                vec = new vector[long long]()
                # Walk the predecessors tree to find our path, we build it up in a C++ vector because we can't know how
                # long it'll be
                p = dest_index
                while p != origin_index:
                    connector = thread_conn[p]
                    p = thread_predecessors[p]
                    vec.push_back(connector)

                for connector in d(vec):
                    thread_cost[connector] = penalty * thread_cost[connector]

                reverse(vec.begin(), vec.end())

                # To prevent runaway algorithms if we find N duplicate routes we should stop
                status = route_set.insert(vec)
                miss_count = miss_count + (not status.second)
                if miss_count > max_misses:
                    break
            else:
                break

    def get_results(self):  # Cython doesn't like this type annotation... -> pa.Table:
        """
        :Returns:
            **route sets** (:obj:`pyarrow.Table`): Returns a table of OD pairs to lists of link IDs for
                each OD pair provided (as columns). Represents paths from ``origin`` to ``destination``.
        """
        if self.results is None:
            raise RuntimeError("Route Choice results not computed yet")

        return self.results.make_table_from_results()

    def get_link_loading(RouteChoiceSet self, cores: int = 0):
        """
        :Returns:
            **link loading results** (:obj:`Dict[str, np.array]`): Returns a dict of demand column names to
                uncompressed link loads.
        """
        if self.ll_results is None:
            raise RuntimeError("Link loading results not computed yet")

        return self.ll_results.link_loading_to_objects(
            self.graph_compressed_id_view,
            cores if cores > 0 else omp_get_max_threads()
        )

    def get_sl_link_loading(RouteChoiceSet self, cores: int = 0):
        """
        :Returns:
            **select link loading results** (:obj:`Dict[str, Dict[str, np.array]]`): Returns a dict of select link set
                names to a dict of demand column names to uncompressed select link loads.
        """
        if self.ll_results is None:
            raise RuntimeError("Link loading results not computed yet")

        return self.ll_results.sl_link_loading_to_objects(
            self.graph_compressed_id_view,
            cores if cores > 0 else omp_get_max_threads()
        )

    def get_sl_od_matrices(RouteChoiceSet self):
        """
        :Returns:
            **select link OD matrix results** (:obj:`Dict[str, Dict[str, scipy.sparse.coo_matrix]]`): Returns a dict of
                select link set names to a dict of demand column names to a sparse OD matrix
        """
        if self.ll_results is None:
            raise RuntimeError("Link loading results not computed yet")

        return self.ll_results.sl_od_matrices_structs_to_objects()



from aequilibrae.matrix.sparse_matrix cimport COO

from cython.operator cimport predecrement as pre_dec
from cython.operator cimport dereference as d
cimport cython

import numpy as np
import itertools

from collections.abc import Hashable

include 'parallel_numpy.pyx'

# See note in route_choice_set.pxd
cdef class LinkLoadingResults:
    def __cinit__(
            self,
            demand: GeneralisedCOODemand,
            select_links: Dict[Hashable, FrozenSet[FrozenSet[int]]],
            num_links: int,
            sl_link_loading: bool,
            threads: int
    ):
        if threads <= 0:
            raise ValueError(f"threads must be positive ({threads})")
        elif num_links <= 0:
            raise ValueError(f"num_links must be positive ({num_links})")

        self.demand = demand
        self.num_links = num_links
        self.od_matrix_objects = None
        self.sl_link_loading = sl_link_loading

        cdef:
            vector[unique_ptr[vector[double]]] *f64_demand_cols
            vector[unique_ptr[vector[float]]] *f32_demand_cols

        # Link loading
        # Allocate the threaded f64 link loading.
        self.f64_link_loading_threaded.reserve(threads)
        for i in range(threads):
            f64_demand_cols = new vector[unique_ptr[vector[double]]]()
            f64_demand_cols.reserve(len(self.demand.f64_names))

            for j in range(len(self.demand.f64_names)):
                f64_demand_cols.emplace_back(new vector[double](self.num_links))

            self.f64_link_loading_threaded.emplace_back(f64_demand_cols)

        # Allocate the threaded f32 link loading.
        self.f32_link_loading_threaded.reserve(threads)
        for i in range(threads):
            f32_demand_cols = new vector[unique_ptr[vector[float]]]()
            f32_demand_cols.reserve(len(self.demand.f32_names))

            for j in range(len(self.demand.f32_names)):
                f32_demand_cols.emplace_back(new vector[float](self.num_links))

            self.f32_link_loading_threaded.emplace_back(f32_demand_cols)

        # self.f64_link_loading and self.f32_link_loading are not allocated here. The objects are initialised to empty
        # vectors but elements are created in self.reduce_link_loading

        # Select link loading
        cdef:
            vector[unique_ptr[unordered_set[long long]]] *select_link_set
            vector[size_t] *select_link_set_length

        # Select link loading sets
        # Coerce the select link sets to their cpp structures ahead of time. We'll be using these a lot and they don't
        # change. We allocate a vector of select link sets. These select link sets a vector representing an OR set,
        # containing a unordered_set of links representing the AND set.

        self.select_link_set_names = select_links.keys()

        self.select_link_sets.reserve(len(select_links))
        self.select_link_set_lengths.reserve(len(select_links))
        for or_set in select_links.values():
            select_link_set = new vector[unique_ptr[unordered_set[long long]]]()
            select_link_set_length = new vector[size_t]()

            select_link_set.reserve(len(or_set))
            select_link_set_length.reserve(len(or_set))

            for and_set in or_set:
                select_link_set.emplace_back(new unordered_set[long long](and_set))
                select_link_set_length.push_back(len(and_set))

            self.select_link_sets.emplace_back(select_link_set)
            self.select_link_set_lengths.emplace_back(select_link_set_length)

        # Select link loading link loads
        cdef:
            vector[unique_ptr[vector[unique_ptr[vector[double]]]]] *f64_sl_select_link_sets
            vector[unique_ptr[vector[unique_ptr[vector[float]]]]] *f32_sl_select_link_sets
            vector[unique_ptr[vector[double]]] *f64_sl_demand_cols
            vector[unique_ptr[vector[float]]] *f32_sl_demand_cols

        if self.sl_link_loading:
            # Allocate f64 thread storage for select link
            self.f64_sl_link_loading_threaded.reserve(threads)
            for i in range(threads):
                f64_sl_select_link_sets = new vector[unique_ptr[vector[unique_ptr[vector[double]]]]]()
                f64_sl_select_link_sets.reserve(self.select_link_sets.size())

                for j in range(self.select_link_sets.size()):
                    f64_sl_demand_cols = new vector[unique_ptr[vector[double]]]()
                    f64_sl_demand_cols.reserve(len(self.demand.f64_names))

                    for k in range(len(self.demand.f64_names)):
                        f64_sl_demand_cols.emplace_back(new vector[double](self.num_links))

                    f64_sl_select_link_sets.emplace_back(f64_sl_demand_cols)

                self.f64_sl_link_loading_threaded.emplace_back(f64_sl_select_link_sets)

            # Allocate f32 thread storage for select link
            self.f32_sl_link_loading_threaded.reserve(threads)
            for i in range(threads):
                f32_sl_select_link_sets = new vector[unique_ptr[vector[unique_ptr[vector[float]]]]]()
                f32_sl_select_link_sets.reserve(self.select_link_sets.size())

                for j in range(self.select_link_sets.size()):
                    f32_sl_demand_cols = new vector[unique_ptr[vector[float]]]()
                    f32_sl_demand_cols.reserve(len(self.demand.f32_names))

                    for k in range(len(self.demand.f32_names)):
                        f32_sl_demand_cols.emplace_back(new vector[float](self.num_links))

                    f32_sl_select_link_sets.emplace_back(f32_sl_demand_cols)

                self.f32_sl_link_loading_threaded.emplace_back(f32_sl_select_link_sets)

        # self.f64_sl_link_loading and self.f32_sl_link_loading are not allocated here. The objects are initialised to
        # empty vectors but elements are created in self.reduce_sl_link_loading

        # Select link loading od matrix
        cdef:
            vector[unique_ptr[vector[COO_f64_struct]]] *f64_sl_od_matrix_sets
            vector[unique_ptr[vector[COO_f32_struct]]] *f32_sl_od_matrix_sets
            vector[COO_f64_struct] *f64_sl_od_matrix_demand_cols
            vector[COO_f32_struct] *f32_sl_od_matrix_demand_cols

        # Allocate f64 thread storage for select link
        self.f64_sl_od_matrix_threaded.reserve(threads)
        for i in range(threads):
            f64_sl_od_matrix_sets = new vector[unique_ptr[vector[COO_f64_struct]]]()
            f64_sl_od_matrix_sets.reserve(self.select_link_sets.size())

            for j in range(self.select_link_sets.size()):
                f64_sl_od_matrix_demand_cols = new vector[COO_f64_struct](len(self.demand.f64_names))

                for k in range(len(self.demand.f64_names)):
                    COO.init_f64_struct(d(f64_sl_od_matrix_demand_cols)[k])

                f64_sl_od_matrix_sets.emplace_back(f64_sl_od_matrix_demand_cols)

            self.f64_sl_od_matrix_threaded.emplace_back(f64_sl_od_matrix_sets)

        # Allocate f32 thread storage for select link
        self.f32_sl_od_matrix_threaded.reserve(threads)
        for i in range(threads):
            f32_sl_od_matrix_sets = new vector[unique_ptr[vector[COO_f32_struct]]]()
            f32_sl_od_matrix_sets.reserve(self.select_link_sets.size())

            for j in range(self.select_link_sets.size()):
                f32_sl_od_matrix_demand_cols = new vector[COO_f32_struct](len(self.demand.f32_names))

                for k in range(len(self.demand.f32_names)):
                    COO.init_f32_struct(d(f32_sl_od_matrix_demand_cols)[k])

                f32_sl_od_matrix_sets.emplace_back(f32_sl_od_matrix_demand_cols)

            self.f32_sl_od_matrix_threaded.emplace_back(f32_sl_od_matrix_sets)

        # self.f64_sl_od_matrix and self.f32_sl_od_matrix are not allocated here. The objects are initialised to
        # empty vectors but elements are created in self.reduce_sl_link_loading

    cdef object link_loading_to_objects(self, long long[:] compressed_id_view, int cores):
        if self.link_loading_objects is None:
            self.link_loading_objects = dict(zip(*self.apply_generic_link_loading(
                self.f64_link_loading, self.f32_link_loading, compressed_id_view, cores
            )))
        return self.link_loading_objects

    cdef object sl_link_loading_to_objects(self, long long[:] compressed_id_view, int cores):
        if not self.sl_link_loading:
            return {}

        if self.sl_link_loading_objects is None:
            results = []
            for i in range(self.select_link_sets.size()):
                results.append(dict(zip(*self.apply_generic_link_loading(
                    d(self.f64_sl_link_loading[i]), d(self.f32_sl_link_loading[i]), compressed_id_view, cores
                ))))
            self.sl_link_loading_objects = dict(zip(self.select_link_set_names, results))
        return self.sl_link_loading_objects

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void link_load_single_route_set(
        LinkLoadingResults self,
        const size_t od_idx,
        const RouteVec_t &route_set,
        const vector[double] &prob_vec,
        const size_t thread_id
    ) noexcept nogil:
        cdef:
            # Cython doesn't allow declaring references to objects outside of function signatures. So we get the raw
            # pointer instead. It is still owned by the unique_ptr.
            vector[unique_ptr[vector[double]]] *f64_ll_cols = self.f64_link_loading_threaded[thread_id].get()
            vector[unique_ptr[vector[float]]] *f32_ll_cols = self.f32_link_loading_threaded[thread_id].get()
            vector[double] *f64_ll
            vector[float] *f32_ll

            double f64_demand, f64_load
            float f32_demand, f32_load
            size_t i, j

        # For each demand column
        for i in range(self.demand.f64.size()):
            # we grab a pointer to the relevant link loading vector and demand value.
            f64_demand = d(self.demand.f64[i])[od_idx]
            f64_ll = d(f64_ll_cols)[i].get()

            # For each route in the route set
            for j in range(route_set.size()):
                # we compute out load,
                f64_load = prob_vec[j] * f64_demand
                if f64_load == 0.0:
                    continue

                # then apply that to every link in the route
                for link in d(route_set[j]):
                    d(f64_ll)[link] = d(f64_ll)[link] + f64_load

        # Then we do it all over again for the floats.
        for i in range(self.demand.f32.size()):
            f32_demand = d(self.demand.f32[i])[od_idx]
            f32_ll = d(f32_ll_cols)[i].get()

            for j in range(route_set.size()):
                f32_load = prob_vec[j] * f32_demand
                if f32_load == 0.0:
                    continue

                for link in d(route_set[j]):
                    d(f32_ll)[link] = d(f32_ll)[link] + f32_load

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void reduce_link_loading(LinkLoadingResults self):
        """
        NOTE: doesn't require the GIL but should NOT be called in a multithreaded environment. Thus the function
        requires the GIL.
        """
        cdef:
            vector[unique_ptr[vector[double]]] *f64_ll_cols
            vector[unique_ptr[vector[float]]] *f32_ll_cols
            vector[double] *f64_ll_result
            vector[float] *f32_ll_result
            vector[double] *f64_ll
            vector[float] *f32_ll

            size_t thread_id, i, j

        # Allocate the result link loads
        self.f64_link_loading.reserve(self.demand.f64.size())
        for i in range(self.demand.f64.size()):
            self.f64_link_loading.emplace_back(new vector[double](self.num_links))

        self.f32_link_loading.reserve(self.demand.f32.size())
        for i in range(self.demand.f32.size()):
            self.f32_link_loading.emplace_back(new vector[float](self.num_links))

        # Here we sum all threads link loads into the results.
        # For each thread
        for thread_id in range(self.f64_link_loading_threaded.size()):
            # we grab the columns
            f64_ll_cols = self.f64_link_loading_threaded[thread_id].get()

            # for each column
            for i in range(d(f64_ll_cols).size()):
                # we get the link loading vector
                f64_ll_result = self.f64_link_loading[i].get()
                f64_ll = d(f64_ll_cols)[i].get()

                # for each link
                for j in range(d(f64_ll).size()):
                    d(f64_ll_result)[j] = d(f64_ll_result)[j] + d(f64_ll)[j]

        # Then we do it all over again for the floats.
        for thread_id in range(self.f32_link_loading_threaded.size()):
            f32_ll_cols = self.f32_link_loading_threaded[thread_id].get()

            for i in range(d(f32_ll_cols).size()):
                f32_ll_result = self.f32_link_loading[i].get()
                f32_ll = d(f32_ll_cols)[i].get()

                for j in range(d(f32_ll).size()):
                    d(f32_ll_result)[j] = d(f32_ll_result)[j] + d(f32_ll)[j]

        # Here we discard all the intermediate results
        self.f64_link_loading_threaded.clear()
        self.f32_link_loading_threaded.clear()
        self.f64_link_loading_threaded.shrink_to_fit()
        self.f32_link_loading_threaded.shrink_to_fit()

    cdef object apply_generic_link_loading(
        LinkLoadingResults self,
        vector[unique_ptr[vector[double]]] &f64_link_loading,
        vector[unique_ptr[vector[float]]] &f32_link_loading,
        long long[:] compressed_id_view,
        int cores
    ):
        cdef:
            vector[double] *f64_ll
            vector[float] *f32_ll
            double[:, :] f64_ll_view
            double[:, :] f64_actual
            float[:, :] f32_ll_view
            float[:, :] f32_actual

        f64_ll_vectors = []
        for i in range(f64_link_loading.size()):
            # Push a single element to the back to prevent an issue caused by links that are not present in the
            # compressed graph causing a OOB access
            f64_ll = f64_link_loading[i].get()
            f64_ll.push_back(0.0)

            # Cast the vector to a memory view
            f64_ll_view = <double [:f64_ll.size(), :1]>f64_ll.data()
            f64_actual = np.zeros((compressed_id_view.shape[0], 1), dtype=np.float64)

            # Assign the compressed link loads to the uncompressed graph
            assign_link_loads_cython(f64_actual, f64_ll_view, compressed_id_view, cores)

            # Delete the memory view object and pop that element off the end.
            del f64_ll_view
            f64_ll.pop_back()

            # Just return the actual link loads. If compressed are required then the memory view will have to be copied
            # before it is deleted. It does not own the vectors data.
            f64_ll_vectors.append(np.asarray(f64_actual).reshape(-1))

        # We do the same for the floats, again
        f32_ll_vectors = []
        for i in range(f32_link_loading.size()):
            f32_ll = f32_link_loading[i].get()
            f32_ll.push_back(0.0)

            f32_ll_view = <float [:f32_ll.size(), :1]>f32_ll.data()
            f32_actual = np.zeros((compressed_id_view.shape[0], 1), dtype=np.float32)

            assign_link_loads_cython(f32_actual, f32_ll_view, compressed_id_view, cores)

            del f32_ll_view
            f32_ll.pop_back()

            f32_ll_vectors.append(np.asarray(f32_actual).reshape(-1))

        return itertools.chain(self.demand.f64_names, self.demand.f32_names), \
            itertools.chain(f64_ll_vectors, f32_ll_vectors)

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    @staticmethod
    cdef bool is_in_select_link_set(
        vector[long long] &route,  # const but cython doesn't allow iteration over const vectors unless an index is used
        const vector[unique_ptr[unordered_set[long long]]] &select_link_set,
        const vector[size_t] &select_link_set_lengths
    ) noexcept nogil:
        """
        Confirms if a given route satisfies the requirements of the select link set.
        **Assumes the route contains unique links**. If a link is present >1 and is in the select link set, this method
        will double count it.
        """

        cdef:
            vector[size_t] link_counts
            long long link

        # We count the number of links within the AND set. Assuming we only see unique links, then if that count goes to
        # 0 we have seen all links present within the AND set. A shortest-path will contain only unique links in
        # paths. This may not be not be true for heuristics, but if the heuristic is that bad that it's traversing a
        # link twice I don't think we should be using it.
        link_counts.insert(link_counts.begin(), select_link_set_lengths.cbegin(), select_link_set_lengths.cend())

        # We iterate over an AND set first in hopes that a whole set is satisfied before checking another. This let's
        # just "short circuit" in a sense
        for i in range(select_link_set.size()):
            for link in route:
                if d(select_link_set[i]).find(link) != d(select_link_set[i]).end():
                    if pre_dec(link_counts[i]) == 0:
                        return True
                else:
                    continue

        return False

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void sl_link_load_single_route_set(
        LinkLoadingResults self,
        const size_t od_idx,
        const RouteVec_t &route_set,
        const vector[double] &prob_vec,
        const long long origin_idx,
        const long long dest_idx,
        const size_t thread_id
    ) noexcept nogil:
        cdef:
            # Cython doesn't allow declaring references to objects outside of function signatures. So we get the raw
            # pointer instead. It is still owned by the unique_ptr.
            vector[unique_ptr[vector[unique_ptr[vector[double]]]]] *f64_ll_sets_cols
            vector[unique_ptr[vector[unique_ptr[vector[float]]]]] *f32_ll_sets_cols
            vector[unique_ptr[vector[COO_f64_struct]]] *f64_od_sets_cols
            vector[unique_ptr[vector[COO_f32_struct]]] *f32_od_sets_cols
            vector[double] *f64_ll
            vector[float] *f32_ll

            double f64_load
            float f32_load
            size_t i, j, k

        if self.sl_link_loading:
            f64_ll_sets_cols = self.f64_sl_link_loading_threaded[thread_id].get()
            f32_ll_sets_cols = self.f32_sl_link_loading_threaded[thread_id].get()
        else:
            f64_ll_sets_cols = NULL
            f32_ll_sets_cols = NULL

        f64_od_sets_cols = self.f64_sl_od_matrix_threaded[thread_id].get()
        f32_od_sets_cols = self.f32_sl_od_matrix_threaded[thread_id].get()

        # For each select link set
        for i in range(self.select_link_sets.size()):
            # for each route in the route set
            for j in range(route_set.size()):
                # if this route satisfies this link set, then we link load
                if not LinkLoadingResults.is_in_select_link_set(
                        d(route_set[j]),
                        d(self.select_link_sets[i]),
                        d(self.select_link_set_lengths[i])
                ):
                    continue

                # For each demand column
                for k in range(self.demand.f64.size()):
                    # we grab a pointer to the relevant link loading vector and compute our load from our demand,
                    f64_load = prob_vec[j] * d(self.demand.f64[k])[od_idx]

                    if f64_load == 0.0:
                        continue

                    COO.f64_struct_append(d(d(f64_od_sets_cols)[i])[k], origin_idx, dest_idx, f64_load)

                    if self.sl_link_loading:
                        f64_ll = d(d(f64_ll_sets_cols)[i])[k].get()
                        # then apply that to every link in the route
                        for link in d(route_set[j]):
                            d(f64_ll)[link] = d(f64_ll)[link] + f64_load

                # then we do it again for f32
                for k in range(self.demand.f32.size()):
                    f32_load = prob_vec[j] * d(self.demand.f32[k])[od_idx]

                    if f32_load == 0.0:
                        continue

                    COO.f32_struct_append(d(d(f32_od_sets_cols)[i])[k], origin_idx, dest_idx, f32_load)

                    if self.sl_link_loading:
                        f32_ll = d(d(f32_ll_sets_cols)[i])[k].get()
                        for link in d(route_set[j]):
                            d(f32_ll)[link] = d(f32_ll)[link] + f32_load

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void reduce_sl_link_loading(LinkLoadingResults self):
        """
        NOTE: doesn't require the GIL but should NOT be called in a multithreaded environment. Thus the function
        requires the GIL.
        """
        cdef:
            vector[unique_ptr[vector[unique_ptr[vector[double]]]]] *f64_sl_ll_sets_cols
            vector[unique_ptr[vector[unique_ptr[vector[float]]]]] *f32_sl_ll_sets_cols
            vector[unique_ptr[vector[double]]] *f64_sl_ll_cols
            vector[unique_ptr[vector[float]]] *f32_sl_ll_cols
            vector[double] *f64_sl_ll_result
            vector[float] *f32_sl_ll_result
            vector[double] *f64_sl_ll
            vector[float] *f32_sl_ll

            size_t thread_id, i, j

        if not self.sl_link_loading:
            return

        # Allocate the result link loads
        self.f64_sl_link_loading.reserve(self.select_link_sets.size())
        for i in range(self.select_link_sets.size()):
            f64_sl_ll_cols = new vector[unique_ptr[vector[double]]]()
            f64_sl_ll_cols.reserve(self.demand.f64.size())

            for j in range(self.demand.f64.size()):
                f64_sl_ll_cols.emplace_back(new vector[double](self.num_links))

            self.f64_sl_link_loading.emplace_back(f64_sl_ll_cols)

        self.f32_sl_link_loading.reserve(self.select_link_sets.size())
        for i in range(self.select_link_sets.size()):
            f32_sl_ll_cols = new vector[unique_ptr[vector[float]]]()
            f32_sl_ll_cols.reserve(self.demand.f32.size())

            for j in range(self.demand.f32.size()):
                f32_sl_ll_cols.emplace_back(new vector[float](self.num_links))

            self.f32_sl_link_loading.emplace_back(f32_sl_ll_cols)

        # Here we sum all threads link loads into the results.
        # For each thread
        for thread_id in range(self.f64_sl_link_loading_threaded.size()):
            # we grab the link set specific one
            f64_sl_ll_sets_cols = self.f64_sl_link_loading_threaded[thread_id].get()

            for i in range(d(f64_sl_ll_sets_cols).size()):
                # we grab the columns
                f64_sl_ll_cols = d(f64_sl_ll_sets_cols)[i].get()

                # for each column
                for j in range(d(f64_sl_ll_cols).size()):
                    # we get the link loading vector
                    f64_sl_ll_result = d(self.f64_sl_link_loading[i])[j].get()
                    f64_sl_ll = d(f64_sl_ll_cols)[j].get()

                    # for each link
                    for k in range(d(f64_sl_ll).size()):
                        d(f64_sl_ll_result)[k] = d(f64_sl_ll_result)[k] + d(f64_sl_ll)[k]

        # Then we do it all over again for the floats.
        for thread_id in range(self.f32_sl_link_loading_threaded.size()):
            f32_sl_ll_sets_cols = self.f32_sl_link_loading_threaded[thread_id].get()

            for i in range(d(f32_sl_ll_sets_cols).size()):
                f32_sl_ll_cols = d(f32_sl_ll_sets_cols)[i].get()

                for j in range(d(f32_sl_ll_cols).size()):
                    f32_sl_ll_result = d(self.f32_sl_link_loading[i])[j].get()
                    f32_sl_ll = d(f32_sl_ll_cols)[j].get()

                    for k in range(d(f32_sl_ll).size()):
                        d(f32_sl_ll_result)[k] = d(f32_sl_ll_result)[k] + d(f32_sl_ll)[k]

        # Here we discard all the intermediate results
        self.f64_sl_link_loading_threaded.clear()
        self.f32_sl_link_loading_threaded.clear()
        self.f64_sl_link_loading_threaded.shrink_to_fit()
        self.f32_sl_link_loading_threaded.shrink_to_fit()

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void reduce_sl_od_matrix(LinkLoadingResults self):
        """
        NOTE: doesn't require the GIL but should NOT be called in a multithreaded environment. Thus the function
        requires the GIL.
        """
        cdef:
            vector[unique_ptr[vector[COO_f64_struct]]] *f64_sl_od_sets_cols
            vector[unique_ptr[vector[COO_f32_struct]]] *f32_sl_od_sets_cols
            vector[COO_f64_struct] *f64_sl_od_cols
            vector[COO_f32_struct] *f32_sl_od_cols
            COO_f64_struct *f64_sl_od_result
            COO_f32_struct *f32_sl_od_result
            COO_f64_struct *f64_sl_od
            COO_f32_struct *f32_sl_od

            size_t thread_id, i, j

        # Allocate f64 thread storage for select link
        self.f64_sl_od_matrix.reserve(self.select_link_sets.size())
        for i in range(self.select_link_sets.size()):
            f64_sl_od_cols = new vector[COO_f64_struct](self.demand.f64.size())

            for j in range(self.demand.f64.size()):
                COO.init_f64_struct(d(f64_sl_od_cols)[j])

            self.f64_sl_od_matrix.emplace_back(f64_sl_od_cols)

        # Allocate f32 thread storage for select link
        self.f32_sl_od_matrix.reserve(self.select_link_sets.size())
        for i in range(self.select_link_sets.size()):
            f32_sl_od_cols = new vector[COO_f32_struct](self.demand.f32.size())

            for j in range(self.demand.f32.size()):
                COO.init_f32_struct(d(f32_sl_od_cols)[j])

            self.f32_sl_od_matrix.emplace_back(f32_sl_od_cols)

        # Here we sum all threads link loads into the results.
        # For each thread
        for thread_id in range(self.f64_sl_od_matrix_threaded.size()):
            # we grab the link set specific one
            f64_sl_od_sets_cols = self.f64_sl_od_matrix_threaded[thread_id].get()

            for i in range(d(f64_sl_od_sets_cols).size()):
                # we grab the columns
                f64_sl_od_cols = d(f64_sl_od_sets_cols)[i].get()

                # for each column
                for j in range(d(f64_sl_od_cols).size()):
                    f64_sl_od_result = &d(self.f64_sl_od_matrix[i])[j]
                    f64_sl_od = &d(f64_sl_od_cols)[j]

                    d(f64_sl_od_result.row).insert(
                        d(f64_sl_od_result.row).end(),
                        d(f64_sl_od.row).cbegin(),
                        d(f64_sl_od.row).cend()
                    )
                    d(f64_sl_od_result.col).insert(
                        d(f64_sl_od_result.col).end(),
                        d(f64_sl_od.col).cbegin(),
                        d(f64_sl_od.col).cend()
                    )
                    d(f64_sl_od_result.f64_data).insert(
                        d(f64_sl_od_result.f64_data).end(),
                        d(f64_sl_od.f64_data).begin(),
                        d(f64_sl_od.f64_data).end()
                    )

        # Then we do it all over again for the floats.
        for thread_id in range(self.f32_sl_od_matrix_threaded.size()):
            # we grab the link set specific one
            f32_sl_od_sets_cols = self.f32_sl_od_matrix_threaded[thread_id].get()

            for i in range(d(f32_sl_od_sets_cols).size()):
                # we grab the columns
                f32_sl_od_cols = d(f32_sl_od_sets_cols)[i].get()

                # for each column
                for j in range(d(f32_sl_od_cols).size()):
                    f32_sl_od_result = &d(self.f32_sl_od_matrix[i])[j]
                    f32_sl_od = &d(f32_sl_od_cols)[j]

                    d(f32_sl_od_result.row).insert(
                        d(f32_sl_od_result.row).end(),
                        d(f32_sl_od.row).cbegin(),
                        d(f32_sl_od.row).cend()
                    )
                    d(f32_sl_od_result.col).insert(
                        d(f32_sl_od_result.col).end(),
                        d(f32_sl_od.col).cbegin(),
                        d(f32_sl_od.col).cend()
                    )
                    d(f32_sl_od_result.f32_data).insert(
                        d(f32_sl_od_result.f32_data).end(),
                        d(f32_sl_od.f32_data).begin(),
                        d(f32_sl_od.f32_data).end()
                    )

        # Here we discard all the intermediate results
        self.f64_sl_od_matrix_threaded.clear()
        self.f32_sl_od_matrix_threaded.clear()
        self.f64_sl_od_matrix_threaded.shrink_to_fit()
        self.f32_sl_od_matrix_threaded.shrink_to_fit()

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef object sl_od_matrices_structs_to_objects(LinkLoadingResults self):
        """
        Convert the stored structs to python COO objects. This transfers the ownership of the memory meaning we need to
        store the newly created objects as this is a one-time operation.
        """
        if self.od_matrix_objects is not None:
            return self.od_matrix_objects

        cdef size_t i, j

        od_matrix_objects = []
        for i in range(self.select_link_sets.size()):
            res = []
            for j in range(self.demand.f64.size()):
                coo = COO.from_f64_struct(d(self.f64_sl_od_matrix[i])[j])
                coo.shape = self.demand.shape
                res.append((self.demand.f64_names[j], coo))

            for j in range(self.demand.f32.size()):
                coo = COO.from_f32_struct(d(self.f32_sl_od_matrix[i])[j])
                coo.shape = self.demand.shape
                res.append((self.demand.f32_names[j], coo))

            od_matrix_objects.append(dict(res))

        self.f64_sl_od_matrix.clear()
        self.f32_sl_od_matrix.clear()
        self.f64_sl_od_matrix.shrink_to_fit()
        self.f32_sl_od_matrix.shrink_to_fit()

        self.od_matrix_objects = dict(zip(self.select_link_set_names, od_matrix_objects))
        return self.od_matrix_objects



import logging
from scipy import sparse
import numpy as np
from aequilibrae.paths.public_transport import HyperpathGenerating


class OptimalStrategies:
    def __init__(self, assig_spec):
        from aequilibrae.paths import TransitAssignment

        self.__assig_spec = assig_spec  # type: TransitAssignment
        self.__logger = assig_spec.logger
        self.__classes = {}
        self.__results = {}
        self.__demand_cols = {}

        for cls in self.__assig_spec.classes:
            cls.results.prepare(cls.graph, cls.matrix)

            self.__results[cls._id] = cls.results
            self.__classes[cls._id] = HyperpathGenerating(
                cls.graph.graph,
                head="a_node",
                tail="b_node",
                trav_time=assig_spec._config["Time field"],
                freq=assig_spec._config["Frequency field"],
            )

            demand = sparse.coo_matrix(cls.matrix.matrix[cls.matrix_core], dtype=np.float64)

            # Since the aeq matrix indexes based on centroids, and the transit graph can make the destinction between origins and destinations,
            # We need to translate the index of the cols in to the destination node_ids for the assignment
            if len(cls.graph.od_node_mapping.columns) == 2:
                o_vert_ids = cls.graph.od_node_mapping.iloc[demand.row]["node_id"].values.astype(np.uint32)
                d_vert_ids = cls.graph.od_node_mapping.iloc[demand.col]["node_id"].values.astype(np.uint32)
            else:
                o_vert_ids = cls.graph.od_node_mapping.iloc[demand.row]["o_node_id"].values.astype(np.uint32)
                d_vert_ids = cls.graph.od_node_mapping.iloc[demand.col]["d_node_id"].values.astype(np.uint32)

            self.__demand_cols[cls._id] = {
                "origin_column": o_vert_ids,
                "destination_column": d_vert_ids,
                "demand_column": demand.data,
            }

    def execute(self):
        for cls_id, hyperpath in self.__classes.items():
            self.__logger.info(f"Executing S&F assignment  for {cls_id}")

            hyperpath.assign(**self.__demand_cols[cls_id], threads=self.__assig_spec.cores)
            self.__results[cls_id].link_loads = hyperpath._edges["volume"].values

    # def run(self, origin=None, destination=None, volume=None):
    #     for cls_id, hyperpath in self.__classes.items():
    #         self.__logger.info(f"Executing S&F single run for {cls_id}")

    #         hyperpath.run(origin, destination, volume)
    #         self.__results[cls_id].link_loads.data["volume"] = hyperpath._edges["volume"].values



from aequilibrae import global_logger

try:
    from aequilibrae.paths.AoN import bpr, delta_bpr, bpr2, delta_bpr2, conical, delta_conical, inrets, delta_inrets
except ImportError as ie:
    global_logger.warning(f"Could not import procedures from the binary. {ie.args}")

all_vdf_functions = ["bpr", "bpr2", "conical", "inrets"]


class VDF:
    """Volume-Delay function

    .. code-block:: python

        >>> from aequilibrae.paths import VDF

        >>> vdf = VDF()
        >>> vdf.functions_available()
        ['bpr', 'bpr2', 'conical', 'inrets']

    """

    def __init__(self):
        self.__dict__["function"] = ""
        self.__dict__["apply_vdf"] = None
        self.__dict__["apply_derivative"] = None

    def __setattr__(self, instance, value) -> None:
        if instance == "function":
            value = value.upper()
            self.__dict__[instance] = value
            if value == "BPR":
                self.__dict__["apply_vdf"] = bpr
                self.__dict__["apply_derivative"] = delta_bpr
            elif value == "BPR2":
                self.__dict__["apply_vdf"] = bpr2
                self.__dict__["apply_derivative"] = delta_bpr2
            elif value == "CONICAL":
                self.__dict__["apply_vdf"] = conical
                self.__dict__["apply_derivative"] = delta_conical
            elif value == "INRETS":
                self.__dict__["apply_vdf"] = inrets
                self.__dict__["apply_derivative"] = delta_inrets
            else:
                raise ValueError("VDF function not available")
        else:
            raise AttributeError("This class only allows you to set the VDF to use")

    def functions_available(self) -> list:
        """returns a list of all functions available"""
        return all_vdf_functions



import yaml

valid_functions = ["EXPO", "GAMMA", "POWER"]
members = ["function", "alpha", "beta"]
model_type = "SyntheticGravityModel"


class SyntheticGravityModel:
    """Simple class object to represent synthetic gravity models"""

    def __init__(self):
        self.function = None
        self.alpha = None
        self.beta = None

    def __setattr__(self, key, value):
        if value is None and key in ["function", "alpha", "beta"]:
            self.__dict__[key] = value
        else:
            if key == "function":
                self.alpha = None
                self.beta = None
                if value not in self.valid_functions:
                    raise ValueError("Function needs to be one of these: " + ", ".join(self.valid_functions))
            else:
                if isinstance(value, float) or isinstance(value, int):
                    if key == "alpha":
                        if self.__dict__.get("function") == "EXPO":
                            raise ValueError("Exponential function does not have an alpha parameter")

                    if key == "beta":
                        if self.function == "POWER":
                            raise ValueError("Inverse power function does not have a beta parameter")
                else:
                    raise ValueError("Parameter needs to be numeric")

            self.__dict__[key] = value

    def __getattr__(self, key):
        if key == "valid_functions":
            return valid_functions
        elif key == "members":
            return members
        elif key == "model_type":
            return model_type
        else:
            return self.__dict__[key]

    def load(self, file_name):
        R"""Loads model from disk. Extension is \*.mod"""
        try:
            with open(file_name, "r") as f:
                model = yaml.safe_load(f)[self.model_type]
                for key, value in model.items():
                    if key in self.members:
                        self.__dict__[key] = value
                    else:
                        raise ValueError("Model has unknown parameters: " + str(key))
        except ValueError as err:
            raise ValueError("File provided is not a valid Synthetic Gravity Model - {}".format(err.__str__())) from err

    def save(self, file_name):
        R"""Saves model to disk in yaml format. Extension is \*.mod"""
        file_name = str(file_name)
        if file_name[-4:].upper() != ".MOD":
            file_name += ".mod"

        model = {model_type: {"function": self.function, "alpha": self.alpha, "beta": self.beta}}

        mod = open(file_name, "w")
        yaml.dump(model, mod, default_flow_style=False)
        mod.flush()
        mod.close()



import csv
import sqlite3
from os.path import join
from typing import List

import pandas as pd

from aequilibrae.transit.transit_elements import Trip


def write_trips(trips: List[Trip], folder_path: str, conn: sqlite3.Connection):
    headers = ["route_id", "service_id", "trip_id", "trip_headsign", "direction_id", "shape_id"]
    data = [[trp.__dict__[hdr] for hdr in headers] for trp in trips]

    all_trips = pd.DataFrame(data, columns=headers)
    all_trips.to_csv(join(folder_path, "trips.txt"), quoting=csv.QUOTE_NONNUMERIC, index=False)

    srvc = pd.read_sql("select service_date from Transit_Agencies where agency_id>1", conn)

    services = all_trips[["service_id"]].drop_duplicates()
    min_date = str(srvc.service_date.min()).replace("-", "")
    max_date = str(srvc.service_date.max()).replace("-", "")
    services = services.assign(start_date=min_date, end_date=max_date)
    services = services.assign(monday=1, tuesday=1, wednesday=1, thursday=1, friday=1, saturday=1, sunday=1)
    services.to_csv(join(folder_path, "calendar.txt"), quoting=csv.QUOTE_NONNUMERIC, index=False)
